# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Dict, Set, Type, NamedTuple

from javalang import tree
from javalang.ast import Node

from aibolit.ast_framework import ASTNodeType


class ASTNodeReference(NamedTuple):
    node_index: int


javalang_to_ast_node_type: Dict[Type[Node], ASTNodeType] = {
    tree.Annotation: ASTNodeType.ANNOTATION,
    tree.AnnotationDeclaration: ASTNodeType.ANNOTATION_DECLARATION,
    tree.AnnotationMethod: ASTNodeType.ANNOTATION_METHOD,
    tree.ArrayCreator: ASTNodeType.ARRAY_CREATOR,
    tree.ArrayInitializer: ASTNodeType.ARRAY_INITIALIZER,
    tree.ArraySelector: ASTNodeType.ARRAY_SELECTOR,
    tree.AssertStatement: ASTNodeType.ASSERT_STATEMENT,
    tree.Assignment: ASTNodeType.ASSIGNMENT,
    tree.BasicType: ASTNodeType.BASIC_TYPE,
    tree.BinaryOperation: ASTNodeType.BINARY_OPERATION,
    tree.BlockStatement: ASTNodeType.BLOCK_STATEMENT,
    tree.BreakStatement: ASTNodeType.BREAK_STATEMENT,
    tree.Cast: ASTNodeType.CAST,
    tree.CatchClause: ASTNodeType.CATCH_CLAUSE,
    tree.CatchClauseParameter: ASTNodeType.CATCH_CLAUSE_PARAMETER,
    tree.ClassCreator: ASTNodeType.CLASS_CREATOR,
    tree.ClassDeclaration: ASTNodeType.CLASS_DECLARATION,
    tree.ClassReference: ASTNodeType.CLASS_REFERENCE,
    tree.CompilationUnit: ASTNodeType.COMPILATION_UNIT,
    tree.ConstantDeclaration: ASTNodeType.CONSTANT_DECLARATION,
    tree.ConstructorDeclaration: ASTNodeType.CONSTRUCTOR_DECLARATION,
    tree.ContinueStatement: ASTNodeType.CONTINUE_STATEMENT,
    tree.Creator: ASTNodeType.CREATOR,
    tree.Declaration: ASTNodeType.DECLARATION,
    tree.Documented: ASTNodeType.DOCUMENTED,
    tree.DoStatement: ASTNodeType.DO_STATEMENT,
    tree.ElementArrayValue: ASTNodeType.ELEMENT_ARRAY_VALUE,
    tree.ElementValuePair: ASTNodeType.ELEMENT_VALUE_PAIR,
    tree.EnhancedForControl: ASTNodeType.ENHANCED_FOR_CONTROL,
    tree.EnumBody: ASTNodeType.ENUM_BODY,
    tree.EnumConstantDeclaration: ASTNodeType.ENUM_CONSTANT_DECLARATION,
    tree.EnumDeclaration: ASTNodeType.ENUM_DECLARATION,
    tree.ExplicitConstructorInvocation: ASTNodeType.EXPLICIT_CONSTRUCTOR_INVOCATION,
    tree.Expression: ASTNodeType.EXPRESSION,
    tree.FieldDeclaration: ASTNodeType.FIELD_DECLARATION,
    tree.ForControl: ASTNodeType.FOR_CONTROL,
    tree.FormalParameter: ASTNodeType.FORMAL_PARAMETER,
    tree.ForStatement: ASTNodeType.FOR_STATEMENT,
    tree.IfStatement: ASTNodeType.IF_STATEMENT,
    tree.Import: ASTNodeType.IMPORT,
    tree.InferredFormalParameter: ASTNodeType.INFERRED_FORMAL_PARAMETER,
    tree.InnerClassCreator: ASTNodeType.INNER_CLASS_CREATOR,
    tree.InterfaceDeclaration: ASTNodeType.INTERFACE_DECLARATION,
    tree.Invocation: ASTNodeType.INVOCATION,
    tree.LambdaExpression: ASTNodeType.LAMBDA_EXPRESSION,
    tree.Literal: ASTNodeType.LITERAL,
    tree.LocalVariableDeclaration: ASTNodeType.LOCAL_VARIABLE_DECLARATION,
    tree.Member: ASTNodeType.MEMBER,
    tree.MemberReference: ASTNodeType.MEMBER_REFERENCE,
    tree.MethodDeclaration: ASTNodeType.METHOD_DECLARATION,
    tree.MethodInvocation: ASTNodeType.METHOD_INVOCATION,
    tree.MethodReference: ASTNodeType.METHOD_REFERENCE,
    tree.PackageDeclaration: ASTNodeType.PACKAGE_DECLARATION,
    tree.Primary: ASTNodeType.PRIMARY,
    tree.ReferenceType: ASTNodeType.REFERENCE_TYPE,
    tree.ReturnStatement: ASTNodeType.RETURN_STATEMENT,
    tree.Statement: ASTNodeType.STATEMENT,
    tree.StatementExpression: ASTNodeType.STATEMENT_EXPRESSION,
    tree.SuperConstructorInvocation: ASTNodeType.SUPER_CONSTRUCTOR_INVOCATION,
    tree.SuperMemberReference: ASTNodeType.SUPER_MEMBER_REFERENCE,
    tree.SuperMethodInvocation: ASTNodeType.SUPER_METHOD_INVOCATION,
    tree.SwitchStatement: ASTNodeType.SWITCH_STATEMENT,
    tree.SwitchStatementCase: ASTNodeType.SWITCH_STATEMENT_CASE,
    tree.SynchronizedStatement: ASTNodeType.SYNCHRONIZED_STATEMENT,
    tree.TernaryExpression: ASTNodeType.TERNARY_EXPRESSION,
    tree.This: ASTNodeType.THIS,
    tree.ThrowStatement: ASTNodeType.THROW_STATEMENT,
    tree.TryResource: ASTNodeType.TRY_RESOURCE,
    tree.TryStatement: ASTNodeType.TRY_STATEMENT,
    tree.Type: ASTNodeType.TYPE,
    tree.TypeArgument: ASTNodeType.TYPE_ARGUMENT,
    tree.TypeDeclaration: ASTNodeType.TYPE_DECLARATION,
    tree.TypeParameter: ASTNodeType.TYPE_PARAMETER,
    tree.VariableDeclaration: ASTNodeType.VARIABLE_DECLARATION,
    tree.VariableDeclarator: ASTNodeType.VARIABLE_DECLARATOR,
    tree.VoidClassReference: ASTNodeType.VOID_CLASS_REFERENCE,
    tree.WhileStatement: ASTNodeType.WHILE_STATEMENT,
}

common_attributes: Set[str] = {'node_type', 'line'}

attributes_by_node_type: Dict[ASTNodeType, Set[str]] = {
    ASTNodeType.ANNOTATION_DECLARATION: {
        "annotations",
        "body",
        "documentation",
        "modifiers",
        "name",
    },
    ASTNodeType.ANNOTATION_METHOD: {
        "annotations",
        "default",
        "dimensions",
        "modifiers",
        "name",
        "return_type",
    },
    ASTNodeType.ANNOTATION: {"element", "name"},
    ASTNodeType.ARRAY_CREATOR: {
        "dimensions",
        "initializer",
        "postfix_operators",
        "prefix_operators",
        "qualifier",
        "selectors",
        "type",
    },
    ASTNodeType.ARRAY_INITIALIZER: {"initializers"},
    ASTNodeType.ARRAY_SELECTOR: {"index"},
    ASTNodeType.ASSERT_STATEMENT: {"condition", "label", "value"},
    ASTNodeType.ASSIGNMENT: {"expressionl", "type", "value"},
    ASTNodeType.BASIC_TYPE: {"dimensions", "name"},
    ASTNodeType.BINARY_OPERATION: {"operandl", "operandr", "operator"},
    ASTNodeType.BLOCK_STATEMENT: {"label", "statements"},
    ASTNodeType.BREAK_STATEMENT: {"goto", "label"},
    ASTNodeType.CAST: {"expression", "type"},
    ASTNodeType.CATCH_CLAUSE_PARAMETER: {"annotations", "name", "modifiers", "types"},
    ASTNodeType.CATCH_CLAUSE: {"block", "label", "parameter"},
    ASTNodeType.CLASS_CREATOR: {
        "arguments",
        "body",
        "constructor_type_arguments",
        "postfix_operators",
        "prefix_operators",
        "qualifier",
        "selectors",
        "type",
    },
    ASTNodeType.CLASS_DECLARATION: {
        "annotations",
        "body",
        "documentation",
        "extends",
        "implements",
        "modifiers",
        "name",
        "type_parameters",
    },
    ASTNodeType.CLASS_REFERENCE: {
        "postfix_operators",
        "prefix_operators",
        "qualifier",
        "selectors",
        "type",
    },
    ASTNodeType.COLLECTION: set(),
    ASTNodeType.COMPILATION_UNIT: {"imports", "package", "types"},
    ASTNodeType.CONSTANT_DECLARATION: {
        "annotations",
        "declarators",
        "documentation",
        "modifiers",
        "type",
    },
    ASTNodeType.CONSTRUCTOR_DECLARATION: {
        "annotations",
        "body",
        "documentation",
        "modifiers",
        "name",
        "parameters",
        "throws",
        "type_parameters",
    },
    ASTNodeType.CONTINUE_STATEMENT: {"goto", "label"},
    ASTNodeType.CREATOR: {
        "postfix_operators",
        "prefix_operators",
        "qualifier",
        "selectors",
        "type",
    },
    ASTNodeType.DECLARATION: {"annotations", "modifiers"},
    ASTNodeType.DO_STATEMENT: {"body", "condition", "label"},
    ASTNodeType.DOCUMENTED: {"documentation"},
    ASTNodeType.ELEMENT_ARRAY_VALUE: {"values"},
    ASTNodeType.ELEMENT_VALUE_PAIR: {"name", "value"},
    ASTNodeType.ENHANCED_FOR_CONTROL: {"iterable", "var"},
    ASTNodeType.ENUM_BODY: {"constants", "declarations"},
    ASTNodeType.ENUM_CONSTANT_DECLARATION: {
        "annotations",
        "arguments",
        "body",
        "documentation",
        "modifiers",
        "name",
    },
    ASTNodeType.ENUM_DECLARATION: {
        "annotations",
        "body",
        "documentation",
        "implements",
        "modifiers",
        "name",
    },
    ASTNodeType.EXPLICIT_CONSTRUCTOR_INVOCATION: {
        "arguments",
        "postfix_operators",
        "prefix_operators",
        "qualifier",
        "selectors",
        "type_arguments",
    },
    ASTNodeType.EXPRESSION: set(),
    ASTNodeType.FIELD_DECLARATION: {
        "annotations",
        "declarators",
        "documentation",
        "modifiers",
        "type",
    },
    ASTNodeType.FOR_CONTROL: {"condition", "init", "update"},
    ASTNodeType.FOR_STATEMENT: {"body", "control", "label"},
    ASTNodeType.FORMAL_PARAMETER: {
        "annotations",
        "modifiers",
        "name",
        "type",
        "varargs",
    },
    ASTNodeType.IF_STATEMENT: {
        "condition",
        "else_statement",
        "label",
        "then_statement",
    },
    ASTNodeType.IMPORT: {"path", "static", "wildcard"},
    ASTNodeType.INFERRED_FORMAL_PARAMETER: {"name"},
    ASTNodeType.INNER_CLASS_CREATOR: {
        "arguments",
        "body",
        "constructor_type_arguments",
        "postfix_operators",
        "prefix_operators",
        "qualifier",
        "selectors",
        "type",
    },
    ASTNodeType.INTERFACE_DECLARATION: {
        "annotations",
        "body",
        "documentation",
        "extends",
        "modifiers",
        "name",
        "type_parameters",
    },
    ASTNodeType.INVOCATION: {
        "arguments",
        "postfix_operators",
        "prefix_operators",
        "qualifier",
        "selectors",
        "type_arguments",
    },
    ASTNodeType.LAMBDA_EXPRESSION: {"body", "parameters"},
    ASTNodeType.LITERAL: {
        "postfix_operators",
        "prefix_operators",
        "qualifier",
        "selectors",
        "value",
    },
    ASTNodeType.LOCAL_VARIABLE_DECLARATION: {
        "annotations",
        "declarators",
        "modifiers",
        "type",
    },
    ASTNodeType.MEMBER_REFERENCE: {
        "member",
        "postfix_operators",
        "prefix_operators",
        "qualifier",
        "selectors",
    },
    ASTNodeType.MEMBER: {"documentation"},
    ASTNodeType.METHOD_DECLARATION: {
        "annotations",
        "body",
        "documentation",
        "modifiers",
        "name",
        "parameters",
        "return_type",
        "throws",
        "type_parameters",
    },
    ASTNodeType.METHOD_INVOCATION: {
        "arguments",
        "member",
        "postfix_operators",
        "prefix_operators",
        "qualifier",
        "selectors",
        "type_arguments",
    },
    ASTNodeType.METHOD_REFERENCE: {"expression", "method", "type_arguments"},
    ASTNodeType.PACKAGE_DECLARATION: {
        "annotations",
        "documentation",
        "modifiers",
        "name",
    },
    ASTNodeType.PRIMARY: {
        "postfix_operators",
        "prefix_operators",
        "qualifier",
        "selectors",
    },
    ASTNodeType.REFERENCE_TYPE: {"arguments", "dimensions", "name", "sub_type"},
    ASTNodeType.RETURN_STATEMENT: {"expression", "label"},
    ASTNodeType.STATEMENT_EXPRESSION: {"expression", "label"},
    ASTNodeType.STATEMENT: {"label"},
    ASTNodeType.STRING: {"string"},
    ASTNodeType.SUPER_CONSTRUCTOR_INVOCATION: {
        "arguments",
        "postfix_operators",
        "prefix_operators",
        "qualifier",
        "selectors",
        "type_arguments",
    },
    ASTNodeType.SUPER_MEMBER_REFERENCE: {
        "member",
        "postfix_operators",
        "prefix_operators",
        "qualifier",
        "selectors",
    },
    ASTNodeType.SUPER_METHOD_INVOCATION: {
        "arguments",
        "member",
        "postfix_operators",
        "prefix_operators",
        "qualifier",
        "selectors",
        "type_arguments",
    },
    ASTNodeType.SWITCH_STATEMENT_CASE: {"case", "statements"},
    ASTNodeType.SWITCH_STATEMENT: {"cases", "expression", "label"},
    ASTNodeType.SYNCHRONIZED_STATEMENT: {"block", "label", "lock"},
    ASTNodeType.TERNARY_EXPRESSION: {"condition", "if_false", "if_true"},
    ASTNodeType.THIS: {
        "postfix_operators",
        "prefix_operators",
        "qualifier",
        "selectors",
    },
    ASTNodeType.THROW_STATEMENT: {"expression", "label"},
    ASTNodeType.TRY_RESOURCE: {"annotations", "name", "modifiers", "type", "value"},
    ASTNodeType.TRY_STATEMENT: {
        "block",
        "catches",
        "finally_block",
        "label",
        "resources",
    },
    ASTNodeType.TYPE_ARGUMENT: {"pattern_type", "type"},
    ASTNodeType.TYPE_DECLARATION: {
        "annotations",
        "body",
        "documentation",
        "modifiers",
        "name",
    },
    ASTNodeType.TYPE_PARAMETER: {"extends", "name"},
    ASTNodeType.TYPE: {"dimensions", "name"},
    ASTNodeType.VARIABLE_DECLARATION: {
        "annotations",
        "declarators",
        "modifiers",
        "type",
    },
    ASTNodeType.VARIABLE_DECLARATOR: {"dimensions", "initializer", "name"},
    ASTNodeType.VOID_CLASS_REFERENCE: {
        "postfix_operators",
        "prefix_operators",
        "qualifier",
        "selectors",
        "type",
    },
    ASTNodeType.WHILE_STATEMENT: {"body", "condition", "label"},
}



import re


class RemoveComments:

    def __init__(self):
        pass

    @staticmethod
    def remove_comments(string):
        # remove all occurrences streamed comments (/*COMMENT */) from string
        string = re.sub(re.compile(r"/\*.*?\*/", re.DOTALL), "",
                        string)
        # remove all occurrence single-line comments (//COMMENT\n ) from string
        string = re.sub(re.compile(r"//.*?\n"), "",
                        string)
        return string




import os
from unittest import TestCase
from aibolit.patterns.non_final_class.non_final_class import NonFinalClass


class TestNonFinalClass(TestCase):

    def test_find_final_class(self):
        path = os.path.dirname(os.path.realpath(__file__)) + '/1.java'
        lines = NonFinalClass().value(path)
        self.assertEqual(lines, [])

    def test_find_non_final_class(self):
        path = os.path.dirname(os.path.realpath(__file__)) + '/2.java'
        lines = NonFinalClass().value(path)
        self.assertEqual(lines, [1])

    def test_abstract_class(self):
        path = os.path.dirname(os.path.realpath(__file__)) + '/3.java'
        lines = NonFinalClass().value(path)
        self.assertEqual(lines, [])



import javalang

import uuid
from collections import defaultdict

from aibolit.utils.ast_builder import build_ast


class MethodChainFind:

    def __init__(self):
        pass

    def traverse_node(self, node, dict_with_chains, uuid_method):
        if not node:
            return dict_with_chains

        for item in node.children:
            if item and (isinstance(item, tuple) or isinstance(item, list)):
                for j in item:
                    if isinstance(j, javalang.tree.MethodInvocation):
                        if not j.qualifier and j.qualifier != '':
                            # it means that there is method chaining
                            dict_with_chains[uuid_method].append([j.position.line, j.member])
                            self.traverse_node(j, dict_with_chains, uuid_method)
                        else:
                            # it means that we have method invocation without chaining like
                            # result.add(field.getName(), column.columnName(), field.getType());
                            new_uuid = uuid.uuid1()
                            dict_with_chains[new_uuid].append([j.position.line, j.member])
                            self.traverse_node(j, dict_with_chains, new_uuid)

                    elif isinstance(j, javalang.tree.MethodDeclaration):
                        self.traverse_node(j, dict_with_chains, str(uuid.uuid1()))

                    elif isinstance(j, javalang.tree.StatementExpression):
                        self.traverse_node(j, dict_with_chains, uuid_method)

                    elif isinstance(j, javalang.tree.This) or isinstance(j, javalang.tree.ClassCreator):
                        self.traverse_node(j, dict_with_chains, str(uuid.uuid1()))
            elif isinstance(item, javalang.tree.ClassCreator):
                self.traverse_node(item, dict_with_chains, uuid_method)

        return dict_with_chains


    # flake8: noqa: C901
    def value(self, filename: str):
        """
        Travers over AST tree finds method chaining. It is searched in a statement
        :param filename:
        :return:
        List of tuples with LineNumber and List of methods names, e.g.
        [[10, 'func1'], [10, 'fun2']], [[23, 'run'], [23, 'start']]]
        """
        tree = build_ast(filename)
        chain_lst = defaultdict(list)
        for _, node in tree.filter(javalang.tree.StatementExpression):
            if isinstance(node.expression, javalang.tree.MethodInvocation):
                children = node.children
                if isinstance(children[1], javalang.tree.MethodInvocation):
                    uuid_first_method = str(uuid.uuid1())
                    chain_lst[uuid_first_method].append([
                        children[1].position.line, children[1].member])
                    self.traverse_node(children[1], chain_lst, uuid_first_method)

        filtered_dict = list(filter(lambda elem: len(elem) > 1, chain_lst.values()))
        return [item[0][0] for item in filtered_dict]



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import List, Type

import javalang

from aibolit.utils.lines import Lines


# mapping between javalang node class names and Java keywords
NODE_KEYWORD_MAP = {
    'SuperMethodInvocation': 'super',
    'WhileStatement': 'while',
    'ForStatement': 'for',
    'TryStatement': 'try',
    'CatchClause': 'catch',
    'SynchronizedStatement': 'synchronized'
}

# list of nodes creating new scope
NEW_SCOPE_NODES = [
    javalang.tree.MethodDeclaration,
    javalang.tree.IfStatement,
    javalang.tree.ForStatement,
    javalang.tree.SwitchStatement,
    javalang.tree.TryStatement,
    javalang.tree.DoStatement,
    javalang.tree.WhileStatement,
    javalang.tree.BlockStatement,
    javalang.tree.CatchClause,
    javalang.tree.SynchronizedStatement
]


class ASTNode:
    def __init__(self, line, method_line, node, scope):
        self.line = line  # node line number in the file
        self.method_line = method_line  # line number where parent method declared
        self.node = node  # javalang AST node object
        self.scope = scope  # ID of scope this node belongs


class JavalangImproved:
    """
    Thi class flattens AST to provie easier interface.
    Deprecated: flatening the AST leads to lose of information and bugs.
                All patterns using it should start traversing the tree manually.
    """

    def __init__(self, filename: str):
        tree, lines = Lines(filename).value()
        self.tree = tree
        self.lines = lines

    def __find_keyword(self, lines, keyword, start):
        '''
        Args:
            lines (List[str]): List of lines from parsed source code file
            keyword (str): keyword to find
            start (int): Line number to start search
        '''
        for i in range(start - 1, len(lines)):
            if keyword in lines[i]:
                return i + 1
        return -1

    def __fix_line_number_if_possible(self, node: javalang.ast.Node, line_n):
        '''
        Try to figure out the "true" line number of AST node in the source file
        '''
        node_class_name = node.__class__.__name__
        if node_class_name not in NODE_KEYWORD_MAP:
            return line_n
        line = self.__find_keyword(
            self.lines,
            NODE_KEYWORD_MAP[node_class_name],
            line_n
        )
        if line == -1:
            return line_n
        return line

    def __tree_to_nodes(
        self,
        tree: javalang.ast.Node,
        line=1,
        parent_method_line=None,
        scope=0
    ) -> List[ASTNode]:
        '''
        Return AST nodes with line numbers sorted by line number

        Args:
            tree (javalang.ast.Node): AST node
            line (int): Supposed line number of processed AST node in the file
            parent_method_line (int): Nearest line number of the method this node located
            scope (int): ID of scope processed AST node located
        '''

        if hasattr(tree, 'position') and tree.position:
            line = tree.position.line

        line = self.__fix_line_number_if_possible(tree, line)
        if type(tree) in NEW_SCOPE_NODES:
            scope += 1

        if type(tree) in [javalang.tree.MethodDeclaration,
                          javalang.tree.ConstructorDeclaration,
                          javalang.tree.LambdaExpression]:
            parent_method_line = line

        res: List[ASTNode] = []
        for child in tree.children:
            nodes_arr = child if isinstance(child, list) else [child]
            for node in nodes_arr:
                if not hasattr(node, 'children'):
                    continue
                left_siblings_line = res[-1].line if len(res) > 0 else line
                res += self.__tree_to_nodes(
                    node,
                    left_siblings_line,
                    parent_method_line,
                    scope
                )

        return [ASTNode(line, parent_method_line, tree, scope)] + res

    def tree_to_nodes(self) -> List[ASTNode]:
        '''Return AST nodes as list with line numbers sorted by line number'''
        nodes = self.__tree_to_nodes(self.tree)
        return sorted(nodes, key=lambda v: v.line)

    def filter(self, ntypes: List[Type[javalang.tree.Node]]) -> List[ASTNode]:
        nodes = self.tree_to_nodes()
        return list(
            filter(lambda v: type(v.node) in ntypes, nodes)
        )

    def get_empty_lines(self) -> List[int]:
        '''Figure out lines that are either empty or multiline statements'''
        lines_with_nodes = self.get_non_empty_lines()
        max_line = max(lines_with_nodes)
        return list(set(range(1, max_line + 1)).difference(lines_with_nodes))

    def get_non_empty_lines(self) -> List[int]:
        '''Figure out file lines that contains statements'''
        return list(map(lambda v: v.line, self.tree_to_nodes()))



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from unittest import TestCase, skip
from pathlib import Path

from aibolit.ast_framework.java_package import JavaPackage
from aibolit.ast_framework import ASTNodeType


@skip('JavaClass is deprecated')
class JavaClassTestCase(TestCase):
    def test_class_name(self):
        for filename, class_names in JavaClassTestCase._java_packages_with_class_names:
            with self.subTest(f'Filename: {filename}'):
                java_package = JavaPackage(Path(__file__).parent.absolute() / filename)
                self.assertEqual(java_package.java_classes.keys(), class_names)

    def test_class_method(self):
        java_package = JavaPackage(Path(__file__).parent.absolute() / 'SimpleClass.java')
        java_class = java_package.java_classes['Simple']
        java_methods = java_class.methods['Increment']
        self.assertEqual(len(java_methods), 1)
        java_method = next(iter(java_methods))
        java_method_node_types = [java_method.get_type(node) for node in java_method.get_nodes()]
        self.assertEqual(java_method_node_types, JavaClassTestCase._java_method_preorder_traversal_types)

    def test_class_field(self):
        java_package = JavaPackage(Path(__file__).parent.absolute() / 'SimpleClass.java')
        java_class = java_package.java_classes['Simple']
        java_field = java_class.fields['x']
        java_field_node_types = [java_field.get_type(node) for node in java_field.get_nodes()]
        self.assertEqual(java_field_node_types, JavaClassTestCase._java_field_preorder_traversal_types)

    _java_packages_with_class_names = [
        ('SimpleClass.java', {'Simple'}),
        ('TwoClasses.java', {'First', 'Second'})
    ]

    _java_method_preorder_traversal_types = [
        ASTNodeType.METHOD_DECLARATION,
        ASTNodeType.COLLECTION,
        ASTNodeType.STRING,
        ASTNodeType.BASIC_TYPE,
        ASTNodeType.STRING,
        ASTNodeType.STRING,
        ASTNodeType.STATEMENT_EXPRESSION,
        ASTNodeType.ASSIGNMENT,
        ASTNodeType.MEMBER_REFERENCE,
        ASTNodeType.STRING,
        ASTNodeType.STRING,
        ASTNodeType.LITERAL,
        ASTNodeType.STRING,
        ASTNodeType.STRING,
        ASTNodeType.RETURN_STATEMENT,
        ASTNodeType.MEMBER_REFERENCE,
        ASTNodeType.STRING,
        ASTNodeType.STRING,
    ]

    _java_field_preorder_traversal_types = [
        ASTNodeType.FIELD_DECLARATION,
        ASTNodeType.COLLECTION,
        ASTNodeType.STRING,
        ASTNodeType.BASIC_TYPE,
        ASTNodeType.STRING,
        ASTNodeType.VARIABLE_DECLARATOR,
        ASTNodeType.STRING,
        ASTNodeType.LITERAL,
        ASTNodeType.STRING,
    ]



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from unittest import TestCase, skip
from pathlib import Path

from typing import Set

from aibolit.ast_framework import ASTNodeType
from aibolit.ast_framework.java_package import JavaPackage


@skip('JavaPackage class is deprecated')
class JavaPackageTestCase(TestCase):

    def test_java_package_name(self):
        for filename, package_name in JavaPackageTestCase._java_packages_with_names:
            with self.subTest(f'Filename: {filename}'):
                java_package = JavaPackage(Path(__file__).parent.absolute() / filename)
                self.assertEqual(java_package.name, package_name)

    def test_classes(self):
        for filename, flatten_classes in JavaPackageTestCase._java_packages_with_classes:
            with self.subTest(f'Filename: {filename}'):
                java_package = JavaPackage(Path(__file__).parent.absolute() / filename)
                found_class_names: Set[str] = set()
                for class_name in java_package.java_classes:
                    with self.subTest(f'Class name: {class_name}'):
                        found_class_names.add(class_name)
                        java_class = java_package.java_classes[class_name]
                        java_class_node_types = [java_class.get_type(node) for node in java_class.get_nodes()]
                        self.assertEqual(java_class_node_types,
                                         flatten_classes[class_name])
                self.assertEqual(found_class_names, flatten_classes.keys())

    _java_packages_with_names = [
        ('SimpleClass.java', '.'),
        ('TwoClasses.java', 'two.classes')
    ]

    _simple_class_preorder_traversal_types = [
        ASTNodeType.CLASS_DECLARATION,
        ASTNodeType.COLLECTION,
        ASTNodeType.STRING,
        ASTNodeType.FIELD_DECLARATION,
        ASTNodeType.COLLECTION,
        ASTNodeType.STRING,
        ASTNodeType.BASIC_TYPE,
        ASTNodeType.STRING,
        ASTNodeType.VARIABLE_DECLARATOR,
        ASTNodeType.STRING,
        ASTNodeType.LITERAL,
        ASTNodeType.STRING,
        ASTNodeType.METHOD_DECLARATION,
        ASTNodeType.COLLECTION,
        ASTNodeType.STRING,
        ASTNodeType.BASIC_TYPE,
        ASTNodeType.STRING,
        ASTNodeType.STRING,
        ASTNodeType.STATEMENT_EXPRESSION,
        ASTNodeType.ASSIGNMENT,
        ASTNodeType.MEMBER_REFERENCE,
        ASTNodeType.STRING,
        ASTNodeType.STRING,
        ASTNodeType.LITERAL,
        ASTNodeType.STRING,
        ASTNodeType.STRING,
        ASTNodeType.RETURN_STATEMENT,
        ASTNodeType.MEMBER_REFERENCE,
        ASTNodeType.STRING,
        ASTNodeType.STRING,
    ]

    _first_class_preorder_traversal_types = [
        ASTNodeType.CLASS_DECLARATION,
        ASTNodeType.COLLECTION,
        ASTNodeType.STRING,
        ASTNodeType.FIELD_DECLARATION,
        ASTNodeType.COLLECTION,
        ASTNodeType.STRING,
        ASTNodeType.BASIC_TYPE,
        ASTNodeType.STRING,
        ASTNodeType.VARIABLE_DECLARATOR,
        ASTNodeType.STRING,
        ASTNodeType.LITERAL,
        ASTNodeType.STRING,
        ASTNodeType.METHOD_DECLARATION,
        ASTNodeType.COLLECTION,
        ASTNodeType.STRING,
        ASTNodeType.BASIC_TYPE,
        ASTNodeType.STRING,
        ASTNodeType.STRING,
        ASTNodeType.STATEMENT_EXPRESSION,
        ASTNodeType.ASSIGNMENT,
        ASTNodeType.MEMBER_REFERENCE,
        ASTNodeType.STRING,
        ASTNodeType.STRING,
        ASTNodeType.LITERAL,
        ASTNodeType.STRING,
        ASTNodeType.STRING,
        ASTNodeType.RETURN_STATEMENT,
        ASTNodeType.MEMBER_REFERENCE,
        ASTNodeType.STRING,
        ASTNodeType.STRING,
    ]

    _second_class_preorder_traversal_types = [
        ASTNodeType.CLASS_DECLARATION,
        ASTNodeType.COLLECTION,
        ASTNodeType.STRING,
        ASTNodeType.FIELD_DECLARATION,
        ASTNodeType.COLLECTION,
        ASTNodeType.STRING,
        ASTNodeType.BASIC_TYPE,
        ASTNodeType.STRING,
        ASTNodeType.VARIABLE_DECLARATOR,
        ASTNodeType.STRING,
        ASTNodeType.LITERAL,
        ASTNodeType.STRING,
        ASTNodeType.FIELD_DECLARATION,
        ASTNodeType.COLLECTION,
        ASTNodeType.STRING,
        ASTNodeType.BASIC_TYPE,
        ASTNodeType.STRING,
        ASTNodeType.VARIABLE_DECLARATOR,
        ASTNodeType.STRING,
        ASTNodeType.LITERAL,
        ASTNodeType.STRING,
        ASTNodeType.METHOD_DECLARATION,
        ASTNodeType.COLLECTION,
        ASTNodeType.STRING,
        ASTNodeType.BASIC_TYPE,
        ASTNodeType.STRING,
        ASTNodeType.STRING,
        ASTNodeType.STATEMENT_EXPRESSION,
        ASTNodeType.ASSIGNMENT,
        ASTNodeType.MEMBER_REFERENCE,
        ASTNodeType.STRING,
        ASTNodeType.STRING,
        ASTNodeType.LITERAL,
        ASTNodeType.STRING,
        ASTNodeType.STRING,
        ASTNodeType.RETURN_STATEMENT,
        ASTNodeType.MEMBER_REFERENCE,
        ASTNodeType.STRING,
        ASTNodeType.STRING,
        ASTNodeType.METHOD_DECLARATION,
        ASTNodeType.COLLECTION,
        ASTNodeType.STRING,
        ASTNodeType.BASIC_TYPE,
        ASTNodeType.STRING,
        ASTNodeType.STRING,
        ASTNodeType.STATEMENT_EXPRESSION,
        ASTNodeType.ASSIGNMENT,
        ASTNodeType.MEMBER_REFERENCE,
        ASTNodeType.STRING,
        ASTNodeType.STRING,
        ASTNodeType.LITERAL,
        ASTNodeType.STRING,
        ASTNodeType.STRING,
        ASTNodeType.RETURN_STATEMENT,
        ASTNodeType.MEMBER_REFERENCE,
        ASTNodeType.STRING,
        ASTNodeType.STRING,
    ]

    _java_packages_with_classes = [
        ('SimpleClass.java', {'Simple': _simple_class_preorder_traversal_types}),
        ('TwoClasses.java', {
            'First': _first_class_preorder_traversal_types,
            'Second': _second_class_preorder_traversal_types,
        })
    ]



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
from unittest import TestCase
from aibolit.patterns.this_finder.this_finder import ThisFinder


class TestFindThis(TestCase):
    cur_dir = os.path.dirname(os.path.realpath(__file__))
    pattern = ThisFinder()

    def test_several(self):
        lines = self.pattern.value(self.cur_dir + '/several.java')
        self.assertEqual(lines, [4, 10, 20])

    def test_simple1(self):
        lines = self.pattern.value(self.cur_dir + '/double_this.java')
        self.assertEqual(lines, [12])

    def test_simple2(self):
        lines = self.pattern.value(self.cur_dir + '/init_block.java')
        self.assertEqual(lines, [])

    def test_simple22(self):
        lines = self.pattern.value(self.cur_dir + '/init_static_block.java')
        self.assertEqual(lines, [])

    def test_simple3(self):
        lines = self.pattern.value(self.cur_dir + '/autocloseable.java')
        self.assertEqual(lines, [4, 14, 31])

    def test_simple4(self):
        lines = self.pattern.value(self.cur_dir + '/one_line_this.java')
        self.assertEqual(lines, [11])

    def test_simple5(self):
        lines = self.pattern.value(self.cur_dir + '/one_line_usage.java')
        self.assertEqual(lines, [12])

    def test_simple6(self):
        lines = self.pattern.value(self.cur_dir + '/super.java')
        self.assertEqual(lines, [12])

    def test_simple7(self):
        lines = self.pattern.value(self.cur_dir + '/holy_moly_constructor.java')
        self.assertEqual(lines, [47])

    def test_simple9(self):
        lines = self.pattern.value(self.cur_dir + '/super_this.java')
        self.assertEqual(lines, [15, 25, 51, 62, 76, 87, 101])

    def test_simple10(self):
        lines = self.pattern.value(self.cur_dir + '/BookmarkEditCmd.java')
        self.assertEqual(lines, [])

    def test_simple11(self):
        lines = self.pattern.value(self.cur_dir + '/ChainedBuffer.java')
        self.assertEqual(lines, [])

    def test_simple12(self):
        lines = self.pattern.value(self.cur_dir + '/CliMethodExtraSections.java')
        self.assertEqual(lines, [])

    def test_simple13(self):
        lines = self.pattern.value(self.cur_dir + '/LengthStringOrdinalSet.java')
        self.assertEqual(lines, [])

    def test_simple14(self):
        lines = self.pattern.value(self.cur_dir + '/LoaderInfoHeader.java')
        self.assertEqual(lines, [])

    def test_simple15(self):
        lines = self.pattern.value(self.cur_dir + '/OmfModuleEnd.java')
        self.assertEqual(lines, [])



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
from unittest import TestCase
from aibolit.patterns.if_return_if_detection.if_detection import CountIfReturn
from pathlib import Path


class TestCountIfReturn(TestCase):
    dir_path = Path(os.path.realpath(__file__)).parent
    countifreturn = CountIfReturn()

    def test_2nd_level_inside(self):
        lines = self.countifreturn.value(Path(self.dir_path, '1.java'))
        self.assertEqual(lines, [6, 10])

    def test_no_return_inside(self):
        lines = self.countifreturn.value(Path(self.dir_path, '2.java'))
        self.assertEqual(lines, [])

    def test_nested_one_goodreturn(self):
        lines = self.countifreturn.value(Path(self.dir_path, '3.java'))
        self.assertEqual(lines, [])

    def test_nested_one_badreturn(self):
        lines = self.countifreturn.value(Path(self.dir_path, '4.java'))
        self.assertEqual(lines, [6])

    def test_withandwithout_returns(self):
        lines = self.countifreturn.value(Path(self.dir_path, '5.java'))
        self.assertEqual(lines, [6, 14, 16, 18])

    def test_nomoreReturn_and_nested(self):
        lines = self.countifreturn.value(Path(self.dir_path, '6.java'))
        self.assertEqual(lines, [10])



import os
from unittest import TestCase
from pathlib import Path

from aibolit.utils.encoding_detector import detect_encoding_of_file


class TestEncodingDetector(TestCase):
    dir_path = Path(os.path.realpath(__file__)).parent
    files_with_encoding = [('ConditionalExpressionCheck.java', 'UTF-8'),
                           ('ExceptionDemo.java', 'GB18030')]

    def test_encoding_detector(self):
        for filename, excepted_encoding in TestEncodingDetector.files_with_encoding:
            with self.subTest():
                actual_encoding = detect_encoding_of_file(Path(self.dir_path, filename))
                self.assertEqual(actual_encoding, excepted_encoding)



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import numpy as np
from unittest import TestCase
from aibolit.model.model import TwoFoldRankingModel


class TestModel(TestCase):

    def test_get_minimum(self):
        ranking_model = TwoFoldRankingModel()
        c1 = np.array([1, 4, 5, 3, 6, 6, 4, 3, 1])
        c2 = np.array([1, 2, 7, 3, 8, 4, 5, 3, -1])
        c3 = np.array([1, 4, 5, 7, 6, 3, 0, -3, 1])
        c, number = ranking_model.get_minimum(c1, c2, c3)
        np.testing.assert_array_equal(c, np.array([1, 2, 5, 3, 6, 3, 0, -3, -1]))
        np.testing.assert_array_equal(number, np.array([0, 1, 0, 0, 0, 2, 2, 2, 1]))



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from enum import Enum, auto

from typing import List, Set


class ScopeStatusFlags(Enum):
    ONLY_VARIABLE_DECLARATIONS_PRESENT = auto()
    INSIDE_VARIABLE_DECLARATION_SUBTREE = auto()
    INSIDE_CALLING_SUPER_CLASS_CONSTRUCTOR_SUBTREE = auto()
    INSIDE_ANNOTATION_SUBTREE = auto()


class ScopeStatus:
    _default_scope_status: Set[ScopeStatusFlags] = \
        {ScopeStatusFlags.ONLY_VARIABLE_DECLARATIONS_PRESENT}

    def __init__(self):
        # Copy _default_scope_status to prevent its modification
        self._scope_stack: List[Set[ScopeStatusFlags]] = \
            [ScopeStatus._default_scope_status.copy()]

    def get_status(self):
        try:
            return self._scope_stack[-1]
        except IndexError:
            raise RuntimeError("No scopes registered.")

    def add_flag(self, flag):
        try:
            self._scope_stack[-1].add(flag)
        except IndexError:
            raise RuntimeError("No scopes registered.")

    def remove_flag(self, flag):
        try:
            self._scope_stack[-1].discard(flag)
        except IndexError:
            raise RuntimeError("No scopes registered.")

    def enter_new_scope(self, new_scope_status=_default_scope_status):
        # Copy new_scope_status to prevent its modification
        self._scope_stack.append(new_scope_status.copy())

    def leave_current_scope(self):
        try:
            self._scope_stack.pop()
        except IndexError:
            raise RuntimeError("No scopes registered.")



import os
from pathlib import Path

from aibolit.metrics.number_variables.numVariables import NumVars as M7
from aibolit.metrics.cognitiveC.cognitive_c import CognitiveComplexity as M4
from aibolit.metrics.entropy.entropy import Entropy as M1
from aibolit.metrics.lcom4.lcom4 import LCOM4 as M5
from aibolit.metrics.maxDiameter.max_diam_of_tree import MaxDiamOfTree as M6
from aibolit.metrics.ncss.ncss import NCSSMetric as M2
from aibolit.metrics.spaces.SpaceCounter import IndentationCounter as M3
from aibolit.patterns.array_as_argument.array_as_argument import ArrayAsArgument as P22
from aibolit.patterns.assert_in_code.assert_in_code import AssertInCode as P1
from aibolit.patterns.assign_null_finder.assign_null_finder import NullAssignment as P28
from aibolit.patterns.classic_setter.classic_setter import ClassicSetter as P2
from aibolit.patterns.empty_rethrow.empty_rethrow import EmptyRethrow as P3
from aibolit.patterns.er_class.er_class import ErClass as P4
from aibolit.patterns.force_type_casting_finder.force_type_casting_finder import ForceTypeCastingFinder as P5
from aibolit.patterns.if_return_if_detection.if_detection import CountIfReturn as P6
from aibolit.patterns.implements_multi.implements_multi import ImplementsMultiFinder as P7
from aibolit.patterns.instanceof.instance_of import InstanceOf as P8
from aibolit.patterns.joined_validation.joined_validation import JoinedValidation as P23
from aibolit.patterns.many_primary_ctors.many_primary_ctors import ManyPrimaryCtors as P9
from aibolit.patterns.method_chaining.method_chaining import MethodChainFind as P10
from aibolit.patterns.multiple_try.multiple_try import MultipleTry as P11
from aibolit.patterns.multiple_while.multiple_while import MultipleWhile as P29
from aibolit.patterns.nested_loop.nested_loop import NestedLoop as P32
from aibolit.patterns.non_final_attribute.non_final_attribute import NonFinalAttribute as P12
from aibolit.patterns.non_final_class.non_final_class import NonFinalClass as P24
from aibolit.patterns.null_check.null_check import NullCheck as P13
from aibolit.patterns.partial_synchronized.partial_synchronized import PartialSync as P14
from aibolit.patterns.private_static_method.private_static_method import PrivateStaticMethod as P25
from aibolit.patterns.protected_method.protected_method import ProtectedMethod as P30
from aibolit.patterns.public_static_method.public_static_method import PublicStaticMethod as P26
from aibolit.patterns.redundant_catch.redundant_catch import RedundantCatch as P15
from aibolit.patterns.return_null.return_null import ReturnNull as P16
from aibolit.patterns.send_null.send_null import SendNull as P31
from aibolit.patterns.string_concat.string_concat import StringConcatFinder as P17
from aibolit.patterns.supermethod.supermethod import SuperMethod as P18
from aibolit.patterns.this_finder.this_finder import ThisFinder as P19
from aibolit.patterns.var_decl_diff.var_decl_diff import VarDeclarationDistance as P20
from aibolit.patterns.var_middle.var_middle import VarMiddle as P21
from aibolit.patterns.var_siblings.var_siblings import VarSiblings as P27


class Singleton(type):
    _instances = {}  # type: ignore

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class Config(metaclass=Singleton):

    @staticmethod
    def home_aibolit_folder():
        return os.environ.get('HOME_AIBOLIT') or '/home/jovyan/aibolit'

    @staticmethod
    def folder_to_save_model_data():
        model_folder = Path(Config().home_aibolit_folder(), 'aibolit', 'binary_files')
        return os.environ.get('SAVE_MODEL_FOLDER') or model_folder

    @staticmethod
    def folder_model_data():
        dir_path = os.path.dirname(os.path.realpath(__file__))
        model_file = Path(Path(dir_path), 'binary_files', 'model.pkl')
        return os.environ.get('HOME_MODEL_FOLDER') or model_file

    @staticmethod
    def dataset_file():
        dataset_path = Path(Config().home_aibolit_folder(), 'scripts', 'target', 'dataset.csv')
        return os.environ.get('HOME_DATASET_CSV') or dataset_path

    @staticmethod
    def get_patterns_config():
        return {
            "patterns": [
                {"name": "Asserts", "code": "P1", "make": lambda: P1()},
                {"name": "Setters", "code": "P2", "make": lambda: P2()},
                {"name": "Empty Rethrow", "code": "P3", "make": lambda: P3()},
                {"name": "Prohibited class name", "code": "P4", "make": lambda: P4()},
                {"name": "Force Type Casting", "code": "P5", "make": lambda: P5()},
                {"name": "Count If Return", "code": "P6", "make": lambda: P6()},
                {"name": "Implements Multi", "code": "P7", "make": lambda: P7()},
                {"name": "Instance of", "code": "P8", "make": lambda: P8()},
                {"name": "Many primary constructors", "code": "P9", "make": lambda: P9()},
                {"name": "Method chain", "code": "P10", "make": lambda: P10()},
                {"name": "Multiple try", "code": "P11", "make": lambda: P11()},
                {"name": "Non final attribute", "code": "P12", "make": lambda: P12()},
                {"name": "Null check", "code": "P13", "make": lambda: P13()},
                {"name": "Partial synchronized", "code": "P14", "make": lambda: P14()},
                {"name": "Redundant catch", "code": "P15", "make": lambda: P15()},
                {"name": "Return null", "code": "P16", "make": lambda: P16()},
                {"name": "String concat", "code": "P17", "make": lambda: P17()},
                {"name": "Super Method", "code": "P18", "make": lambda: P18()},
                {"name": "This in constructor", "code": "P19", "make": lambda: P19()},
                {
                    "name": "Var declaration distance for 5 lines",
                    "code": "P20_5",
                    "make": lambda: P20(5)
                },
                {
                    "name": "Var declaration distance for 7 lines",
                    "code": "P20_7",
                    "make": lambda: P20(7)
                },
                {
                    "name": "Var declaration distance for 11 lines",
                    "code": "P20_11",
                    "make": lambda: P20(11)
                },
                {"name": "Var in the middle", "code": "P21", "make": lambda: P21()},
                {"name": "Array as function argument", "code": "P22", "make": lambda: P22()},
                {"name": "Joined validation", "code": "P23", "make": lambda: P23()},
                {"name": "Non final class", "code": "P24", "make": lambda: P24()},
                {"name": "Private static method", "code": "P25", "make": lambda: P25()},
                {"name": "Public static method", "code": "P26", "make": lambda: P26()},
                {"name": "Var siblings", "code": "P27", "make": lambda: P27()},
                {"name": "Null Assignment", "code": "P28", "make": lambda: P28()},
                {"name": "Multiple While", "code": "P29", "make": lambda: P29()},
                {"name": "Protected Method", "code": "P30", "make": lambda: P30()},
                {"name": "Send Null", "code": "P31", "make": lambda: P31()},
                {"name": "Nested Loop", "code": "P32", "make": lambda: P32()}
            ],
            "metrics": [
                {"name": "Entropy", "code": "M1", "make": lambda: M1()},
                {"name": "NCSS lightweight", "code": "M2", "make": lambda: M2()},
                {
                    "name": "Indentation counter: Right total variance",
                    "code": "M3_1",
                    "make": lambda: M3(right_var=True)
                },
                {
                    "name": "Indentation counter: Left total variance",
                    "code": "M3_2",
                    "make": lambda: M3(left_var=True)
                },
                {
                    "name": "Indentation counter: Right max variance",
                    "code": "M3_3",
                    "make": lambda: M3(max_right=True)
                },
                {
                    "name": "Indentation counter: Left max variance",
                    "code": "M3_4",
                    "make": lambda: M3(max_left=True)
                },
                {"name": "Cognitive Complexity", "code": "M4", "make": lambda: M4()},
                {"name": "LCOM4", "code": "M5", "make": lambda: M5()},
                {"name": "Max diameter of AST", "code": "M6", "make": lambda: M6()},
                {"name": "Number of variables", "code": "M7", "make": lambda: M7()}
            ],
            "target": {

            },
            "patterns_exclude": [
                "P27",  # empty implementation
            ],
            "metrics_exclude": []
        }



import os.path
from pathlib import Path
from unittest import TestCase

from aibolit.patterns.assert_in_code.assert_in_code import AssertInCode


class TestAssertInCode(TestCase):
    cur_file_dir = Path(os.path.realpath(__file__)).parent

    def test_assert_in_code(self):
        file = Path(self.cur_file_dir, 'Book.java')
        self.assertEqual(AssertInCode().value(file), [3])



import os.path
from pathlib import Path
from unittest import TestCase
import unittest
from aibolit.patterns.bidirect_index.bidirect_index import BidirectIndex


@unittest.skip('Not implemented')
class TestBidirectIndex(TestCase):
    dir_path = Path(os.path.realpath(__file__)).parent

    def test_bidirect_index_increase_decrease(self):
        self.assertEqual(
            BidirectIndex().value(Path(self.dir_path, 'BidirectIndexIncreaseDecrease.java')),
            [3],
            'Could not find bidirect index when index increased and then decreased'
        )

    def test_bidirect_index_decrease_increase(self):
        self.assertEqual(
            BidirectIndex().value(Path(self.dir_path, 'BidirectIndexDecreaseIncrease.java')),
            [3],
            'Could not find bidirect index when index decreased and then increased'
        )

    def test_bidirect_index_increase_decrease_assignment(self):
        self.assertEqual(
            BidirectIndex().value(Path(self.dir_path, 'BidirectIndexIncreaseDecreaseAssignment.java')),
            [3],
            'Could not find bidirect index when index increased and then decreased with assignment'
        )

    def test_bidirect_index_increase_assignment_decrease(self):
        self.assertEqual(
            BidirectIndex().value(Path(self.dir_path, 'BidirectIndexIncreaseAssignmentDecrease.java')),
            [3],
            'Could not find bidirect index when index increased with assignment and then decreased'
        )

    def test_bidirect_index_increase_assignment_decrease_assignment(self):
        self.assertEqual(
            BidirectIndex().value(Path(self.dir_path, 'BidirectIndexIncreaseAssignmentDecreaseAssignment.java')),
            [3],
            'Could not find bidirect index when index increased with assignment and then decreased with assignment'
        )

    def test_bidirect_index_hidden_scope_true(self):
        self.assertEqual(
            BidirectIndex().value(Path(self.dir_path, 'BidirectIndexHiddenScope.java')),
            [0],
            'Could not find bidirec index when scope is hidden'
        )

    def test_bidirect_index_outsider(self):
        self.assertEqual(
            BidirectIndex().value(Path(self.dir_path, 'BidirectIndexOutsider.java')),
            [10],
            'Could not find bidirec index when index is ot of loop'
        )



import javalang
from aibolit.utils.ast_builder import build_ast

from typing import List, Type, Any


class CountNumberOfLeaves:
    '''
    Returns number of leaves in class (sum by each method).
    input: file_path
    output: sum of leaves in class by each method
    '''
    def __init__(self):
        pass

    def countLeaves(self, root: Type) -> int:
        # forming the same data type for each object
        root_arr = root if isinstance(root, List) else [root]
        leaves = 0
        not_count: List[Any] = [None, '', set()]

        # traverse through all childs of object. If there is no child, hence we faced leaf
        for node in root_arr:

            if not hasattr(node, 'children') and node not in not_count:
                return leaves + 1

            elif node in not_count:
                continue

            for each_child in node.children:
                leaves += self.countLeaves(each_child)

        return leaves

    def value(self, filename: str) -> int:

        tree = build_ast(filename)
        traversed = []
        for _, class_body in tree.filter(javalang.tree.ClassDeclaration):
            for each_object in class_body.body:
                if isinstance(each_object, javalang.tree.MethodDeclaration):
                    traversed.append(self.countLeaves(each_object.body))

        return sum(traversed)



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from enum import Enum, auto


class ASTNodeType(Enum):
    ANNOTATION = auto()
    ANNOTATION_DECLARATION = auto()
    ANNOTATION_METHOD = auto()
    ARRAY_CREATOR = auto()
    ARRAY_INITIALIZER = auto()
    ARRAY_SELECTOR = auto()
    ASSERT_STATEMENT = auto()
    ASSIGNMENT = auto()
    BASIC_TYPE = auto()
    BINARY_OPERATION = auto()
    BLOCK_STATEMENT = auto()
    BREAK_STATEMENT = auto()
    CAST = auto()
    CATCH_CLAUSE = auto()
    CATCH_CLAUSE_PARAMETER = auto()
    CLASS_CREATOR = auto()
    CLASS_DECLARATION = auto()
    CLASS_REFERENCE = auto()
    COLLECTION = auto()  # Custom type, represent set (as a node) in AST
    COMPILATION_UNIT = auto()
    CONSTANT_DECLARATION = auto()
    CONSTRUCTOR_DECLARATION = auto()
    CONTINUE_STATEMENT = auto()
    CREATOR = auto()
    DECLARATION = auto()
    DO_STATEMENT = auto()
    DOCUMENTED = auto()
    ELEMENT_ARRAY_VALUE = auto()
    ELEMENT_VALUE_PAIR = auto()
    ENHANCED_FOR_CONTROL = auto()
    ENUM_BODY = auto()
    ENUM_CONSTANT_DECLARATION = auto()
    ENUM_DECLARATION = auto()
    EXPLICIT_CONSTRUCTOR_INVOCATION = auto()
    EXPRESSION = auto()
    FIELD_DECLARATION = auto()
    FOR_CONTROL = auto()
    FOR_STATEMENT = auto()
    FORMAL_PARAMETER = auto()
    IF_STATEMENT = auto()
    IMPORT = auto()
    INFERRED_FORMAL_PARAMETER = auto()
    INNER_CLASS_CREATOR = auto()
    INTERFACE_DECLARATION = auto()
    INVOCATION = auto()
    LAMBDA_EXPRESSION = auto()
    LITERAL = auto()
    LOCAL_VARIABLE_DECLARATION = auto()
    MEMBER = auto()
    MEMBER_REFERENCE = auto()
    METHOD_DECLARATION = auto()
    METHOD_INVOCATION = auto()
    METHOD_REFERENCE = auto()
    PACKAGE_DECLARATION = auto()
    PRIMARY = auto()
    REFERENCE_TYPE = auto()
    RETURN_STATEMENT = auto()
    STATEMENT = auto()
    STATEMENT_EXPRESSION = auto()
    STRING = auto()  # Custom type, represent just string in AST
    SUPER_CONSTRUCTOR_INVOCATION = auto()
    SUPER_MEMBER_REFERENCE = auto()
    SUPER_METHOD_INVOCATION = auto()
    SWITCH_STATEMENT = auto()
    SWITCH_STATEMENT_CASE = auto()
    SYNCHRONIZED_STATEMENT = auto()
    TERNARY_EXPRESSION = auto()
    THIS = auto()
    THROW_STATEMENT = auto()
    TRY_RESOURCE = auto()
    TRY_STATEMENT = auto()
    TYPE = auto()
    TYPE_ARGUMENT = auto()
    TYPE_DECLARATION = auto()
    TYPE_PARAMETER = auto()
    UNKNOWN = auto()  # Custom type, used only in parsing javalang AST
    VARIABLE_DECLARATION = auto()
    VARIABLE_DECLARATOR = auto()
    VOID_CLASS_REFERENCE = auto()
    WHILE_STATEMENT = auto()




import os
from unittest import TestCase
from aibolit.patterns.classic_setter.classic_setter import ClassicSetter
from pathlib import Path


class TestSetter(TestCase):
    dir_path = Path(os.path.realpath(__file__)).parent
    setter_finder = ClassicSetter()

    def test_one_valid_patterns(self):
        lines = self.setter_finder.value(Path(self.dir_path, 'BaseKeyframeAnimation.java'))
        self.assertEqual(lines, [40])

    def test_four_setter_patterns(self):
        lines = self.setter_finder.value(Path(self.dir_path, 'Configuration.java'))
        self.assertEqual(lines, [1236, 1240, 3783, 3819])

    def test_another_setter_patterns(self):
        lines = self.setter_finder.value(Path(self.dir_path, 'SequenceFile.java'))
        self.assertEqual(lines, [2849, 2855, 2861, 3127])



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
from unittest import TestCase
from aibolit.patterns.redundant_catch.redundant_catch import RedundantCatch


class RedundantCatchTest(TestCase):

    def test_simple(self):
        pattern = RedundantCatch()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/Simple.java')
        self.assertEqual(lines, [3])

    def test_both_catches(self):
        pattern = RedundantCatch()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/BothCatches.java')
        self.assertEqual(lines, [3])

    def test_fake(self):
        pattern = RedundantCatch()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/TrickyFake.java')
        self.assertEqual(lines, [])

    def test_try_inside_anonymous(self):
        pattern = RedundantCatch()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/TryInsideAnonymous.java')
        self.assertEqual(lines, [6, 14])

    def test_multiple_catch(self):
        pattern = RedundantCatch()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/MultipleCatch.java')
        self.assertEqual(lines, [3])

    def test_sequential_catch(self):
        pattern = RedundantCatch()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/SequentialCatch.java')
        self.assertEqual(lines, [3])

    def test_sequential_catch_try(self):
        pattern = RedundantCatch()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/SequentialCatchTry.java')
        self.assertEqual(lines, [3, 10])

    def test_try_inside_catch(self):
        pattern = RedundantCatch()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/TryInsideCatch.java')
        self.assertEqual(lines, [7])

    def test_try_inside_finally(self):
        pattern = RedundantCatch()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/TryInsideFinally.java')
        self.assertEqual(lines, [8])

    def test_try_inside_try(self):
        pattern = RedundantCatch()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/TryInsideTry.java')
        self.assertEqual(lines, [5])

    def test_catch_with_functions(self):
        pattern = RedundantCatch()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/CatchWithFunctions.java')
        self.assertEqual(lines, [6])

    def test_catch_with_similar_name(self):
        pattern = RedundantCatch()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/NotThrow.java')
        self.assertEqual(lines, [256])

    def test_try_without_throws(self):
        pattern = RedundantCatch()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/ExcelReader.java')
        self.assertEqual(lines, [])

    def test_try_in_constructor(self):
        pattern = RedundantCatch()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/ExcelAnalyserImpl.java')
        self.assertEqual(lines, [43])

    def test_fake_try_in_lambda(self):
        """
        If function has throws, the pattern shouldn't be recognized
        if the same exception is caught in anonymous lambda
        """
        pattern = RedundantCatch()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/Cache.java')
        self.assertEqual(lines, [])



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import re
from typing import List

from aibolit.utils.encoding_detector import read_text_with_autodetected_encoding


class NullAssignment:
    def value(self, filename: str) -> List:
        source_code = read_text_with_autodetected_encoding(filename)
        pattern = r'[^=!><]=(\s)*null(\s)*;'
        return [lineIndex + 1 for lineIndex, line in
                enumerate(source_code.split('\n')) if re.search(pattern, line)]



from javalang.ast import Node
from javalang.tree import CompilationUnit
from typing import List, Tuple

from javalang.parse import parse

from aibolit.utils.encoding_detector import read_text_with_autodetected_encoding


class Lines:
    """
    Return the lines for some AST
    Deprecated: This class is used ony once by JavalangImproved
                and does not provide any complex functionality, so should be removed.
    """
    def __init__(self, filename: str):
        source_code = read_text_with_autodetected_encoding(filename)

        self._lines = source_code.splitlines(keepends=True)
        self._tree: CompilationUnit = parse(source_code)

    def value(self) -> Tuple[Node, List[str]]:
        return self._tree, self._lines



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import argparse
import os
from hashlib import md5
from pathlib import Path
from unittest import TestCase

import javalang

from aibolit.config import Config

from aibolit.__main__ import list_dir, calculate_patterns_and_metrics, \
    create_xml_tree, create_text, format_converter_for_pattern, find_start_and_end_lines, \
    find_annotation_by_node_type, add_pattern_if_ignored


class TestRecommendPipeline(TestCase):

    def __init__(self, *args, **kwargs):
        super(TestRecommendPipeline, self).__init__(*args, **kwargs)
        self.cur_file_dir = Path(os.path.realpath(__file__)).parent
        self.config = Config.get_patterns_config()

    def __create_mock_input(self):
        item = {
            'filename': '1.java',
            'ncss': 100,
            'results': [
                {'pattern_code': 'P23',
                 'pattern_name': 'Some patterns name',
                 'code_lines': [1, 2, 4],
                 'importance': 0.10
                 }
            ]
        }
        another_item = {
            'filename': 'hdd/home/jardani_jovonovich/John_wick.java',
            'ncss': 50,
            'results': [
                {'pattern_code': 'P2',
                 'pattern_name': 'Somebody please get this man a gun',
                 'code_lines': [10, 100, 15000],
                 'importance': 5.67
                 },
                {'pattern_code': 'P4',
                 'pattern_name': 'New item',
                 'code_lines': [5, 6],
                 'importance': 5.67
                 }
            ]
        }
        error_file = {
            'error_string': "Error occured",
            'filename': 'hdd/home/Error.java',
            'results': []
        }
        mock_input = [item, another_item, error_file]
        return mock_input

    def __suppress_argparse_mock(self):
        argparse_mock = argparse.ArgumentParser()
        argparse_mock.add_argument(
            '--suppress',
            default=[],
            nargs="*",
        )
        argparse_mock.suppress = []
        return argparse_mock

    def __create_input_for_xml(self):
        return [
            {'filename': 'D:\\target\\0001\\fast\\Configuration.java',
             'ncss': 100,
             'results': [
                 {'code_lines': [294, 391],
                  'pattern_code': 'P13',
                  'pattern_name': 'Null check',
                  'importance': 30.95612931128819},
                 {'code_lines': [235, 240],
                  'pattern_code': 'P12',
                  'pattern_name': 'Non final attribute',
                  'importance': 17.89671525822768}
             ]},
            {'filename': 'D:\\target\\0001\\fast\\Error.java',
             'results': [],
             'ncss': 0,
             'error_string': "Smth happened"
             },
            {'filename': 'D:\\target\\0001\\fast\\Another.java',
             'ncss': 50,
             'results': [
                 {'code_lines': [23, 2],
                  'pattern_code': 'P13',
                  'pattern_name': 'Null check',
                  'importance': 10.95},
                 {'code_lines': [235, 240],
                  'pattern_code': 'P12',
                  'pattern_name': 'Non final attribute',
                  'importance': 0.23}
             ]}
        ]

    def __create_mock_cmd(self):
        return [
            '/mnt/d/git/aibolit/aibolit/__main__.py',
            'recommend',
            '--folder=/mnt/d/target/0001/fast',
            '--format=compact',
            '--full'
        ]

    def test_calculate_patterns_and_metrics(self):
        args = self.__suppress_argparse_mock()
        file = Path(self.cur_file_dir, 'folder/LottieImageAsset.java')
        input_params, code_lines_dict, error_string = calculate_patterns_and_metrics(file, args)
        val = code_lines_dict['P2']
        self.assertNotEqual(val, 0)
        val = code_lines_dict['P24']
        self.assertNotEqual(val, 0)

    def test_calculate_patterns_and_metrics_wih_suppress(self):
        args = self.__suppress_argparse_mock()
        args.suppress = 'P2'
        file = Path(self.cur_file_dir, 'folder/LottieImageAsset.java')
        input_params, code_lines_dict, error_string = calculate_patterns_and_metrics(file, args)
        val = code_lines_dict['P2']
        self.assertEqual(val, 0)
        val = code_lines_dict['P24']
        self.assertNotEqual(val, 0)

    def test_list_dir_no_java_files(self):
        found_files = []
        file = Path(self.cur_file_dir, 'no_java_files')
        self.assertEqual(list_dir(file, found_files), found_files)

    def test_list_dir(self):
        file = Path(self.cur_file_dir, 'folder')
        found_files = []
        list_dir(file, found_files)
        resuls = {'KeyframeParser.java', 'Metadata.java', 'LottieImageAsset.java'}
        filenames = set([Path(x).name for x in found_files])
        self.assertEqual(filenames, resuls)

    def test_xml_create_full_report(self):
        mock_input = self.__create_input_for_xml()
        mock_cmd = self.__create_mock_cmd()
        create_xml_tree(mock_input, full_report=True, cmd=mock_cmd, exit_code=0)

    def test_xml_empty_results(self):
        mock_cmd = self.__create_mock_cmd()
        create_xml_tree([], full_report=True, cmd=mock_cmd, exit_code=0)

    def test_text_format(self):
        mock_input = self.__create_mock_input()
        new_mock = format_converter_for_pattern(mock_input)
        text = create_text(new_mock, full_report=True)
        md5_hash = md5('\n'.join(text).encode('utf-8'))
        self.assertEqual(md5_hash.hexdigest(), '5efdc7930a66874e9d0c7dcfef734687')

    def test_empty_lines_format(self):
        new_mock = format_converter_for_pattern([])
        text = create_text(new_mock, full_report=True)
        md5_hash = md5('\n'.join(text).encode('utf-8'))
        self.assertEqual(md5_hash.hexdigest(), 'bc22beda46ca18267a677eb32361a2aa')

    def test_text_format_sort_by_code_line(self):
        mock_input = self.__create_mock_input()
        new_mock = format_converter_for_pattern(mock_input, 'code_line')
        text = create_text(new_mock, full_report=True)
        md5_hash = md5('\n'.join(text).encode('utf-8'))
        self.assertEqual(md5_hash.hexdigest(), '1324e129e6badbfb6e10f742667023ae')

    def test_find_start_end_line_function(self):
        # Check start and end line for MethodDeclaration,
        # find_start_and_end_lines is used for functions only at the moment
        file = Path(self.cur_file_dir, 'start_end/LottieImageAsset.java')
        with open(file, 'r', encoding='utf-8') as f:
            tree = javalang.parse.parse(f.read())
            method = list(tree.filter(javalang.tree.MethodDeclaration))[0][1]
            start, end = find_start_and_end_lines(method)
            self.assertEqual(start, 32)
            self.assertEqual(end, 59)

    def test_find_start_end_line_empty_function(self):
        # Check start and end line for MethodDeclaration,
        # find_start_and_end_lines is used for functions only at the moment
        file = Path(self.cur_file_dir, 'start_end/EmptyFunction.java')
        with open(file, 'r', encoding='utf-8') as f:
            tree = javalang.parse.parse(f.read())
            method = list(tree.filter(javalang.tree.MethodDeclaration))[0][1]
            start, end = find_start_and_end_lines(method)
            self.assertEqual(start, 32)
            self.assertEqual(end, 32)

    def test_find_start_end_line_empty_function_with_one_line(self):
        file = Path(self.cur_file_dir, 'start_end/OneLineFunction.java')
        with open(file, 'r', encoding='utf-8') as f:
            tree = javalang.parse.parse(f.read())
            method = list(tree.filter(javalang.tree.MethodDeclaration))[0][1]
            start, end = find_start_and_end_lines(method)
            self.assertEqual(start, 32)
            self.assertEqual(end, 32)

    def test_find_start_end_line_empty_function_with_lambda(self):
        file = Path(self.cur_file_dir, 'start_end/Lambda.java')
        with open(file, 'r', encoding='utf-8') as f:
            tree = javalang.parse.parse(f.read())
            method = list(tree.filter(javalang.tree.MethodDeclaration))[0][1]
            start, end = find_start_and_end_lines(method)
            self.assertEqual(start, 32)
            self.assertEqual(end, 43)

    def test_find_start_end_line_empty_function_with_anonymous_class(self):
        file = Path(self.cur_file_dir, 'start_end/AnonymousClass.java')
        with open(file, 'r', encoding='utf-8') as f:
            tree = javalang.parse.parse(f.read())
            method = list(tree.filter(javalang.tree.MethodDeclaration))[0][1]
            start, end = find_start_and_end_lines(method)
            self.assertEqual(start, 32)
            self.assertEqual(end, 41)

    def test_find_start_end_line_empty_function_in_anonymous_class(self):
        file = Path(self.cur_file_dir, 'start_end/AnonymousClass.java')
        with open(file, 'r', encoding='utf-8') as f:
            tree = javalang.parse.parse(f.read())
            method = list(tree.filter(javalang.tree.MethodDeclaration))[1][1]
            start, end = find_start_and_end_lines(method)
            self.assertEqual(start, 37)
            self.assertEqual(end, 38)

    def test_find_start_end_line_in_function2(self):
        file = Path(self.cur_file_dir, 'start_end/UpDirective.java')
        with open(file, 'r', encoding='utf-8') as f:
            tree = javalang.parse.parse(f.read())
            method = list(tree.filter(javalang.tree.MethodDeclaration))[0][1]
            start, end = find_start_and_end_lines(method)
            self.assertEqual(start, 49)
            self.assertEqual(end, 65)

    def test_find_start_end_line_in_class(self):
        file = Path(self.cur_file_dir, 'start_end/LottieImageAsset.java')
        with open(file, 'r', encoding='utf-8') as f:
            tree = javalang.parse.parse(f.read())
            method = list(tree.filter(javalang.tree.ClassDeclaration))[0][1]
            start, end = find_start_and_end_lines(method)
            self.assertEqual(start, 11)
            self.assertEqual(end, 59)

    def test_find_annotation_by_class_declaration(self):
        file = Path(self.cur_file_dir, 'annotations/ClassAnnotations.java')
        with open(file, 'r', encoding='utf-8') as f:
            tree = javalang.parse.parse(f.read())
            classes_with_annonations = find_annotation_by_node_type(tree, javalang.tree.ClassDeclaration)
            pattern_found = list(classes_with_annonations.values())[0][0]
            self.assertEqual(pattern_found, 'P11')

    def test_find_mutiple_annotations_by_class_declaration(self):
        file = Path(self.cur_file_dir, 'annotations/MutipleAnnotations.java')
        with open(file, 'r', encoding='utf-8') as f:
            tree = javalang.parse.parse(f.read())
            classes_with_annonations = find_annotation_by_node_type(tree, javalang.tree.ClassDeclaration)
            patterns_found = list(classes_with_annonations.values())[0]
            self.assertEqual(patterns_found, ['P23', 'P11'])

    def test_find_annotation_method_declaration(self):
        file = Path(self.cur_file_dir, 'annotations/MutipleAnnotations.java')
        with open(file, 'r', encoding='utf-8') as f:
            tree = javalang.parse.parse(f.read())
            functions_with_annotations = find_annotation_by_node_type(tree, javalang.tree.MethodDeclaration)
            patterns_found_with_functions = [(x.name, y) for x, y in functions_with_annotations.items()]
            self.assertEqual(patterns_found_with_functions, [('set', ['P23']), ('getStreamReader', ['P23', 'P22'])])

    def test_find_annotation_by_field_declaration(self):
        file = Path(self.cur_file_dir, 'annotations/MutipleAnnotations.java')
        with open(file, 'r', encoding='utf-8') as f:
            tree = javalang.parse.parse(f.read())
            fields_with_annotations = find_annotation_by_node_type(tree, javalang.tree.FieldDeclaration)
            patterns_found_with_fields = list(fields_with_annotations.values())
            self.assertEqual(patterns_found_with_fields, [['P23'], ['P23', 'P22']])

    def test_pattern_ignore(self):
        pattern_item = {'code_lines': [20],
                        'pattern_code': 'P13',
                        'pattern_name': 'Null check',
                        'importance': 30.95612931128819}
        results = []
        pattern_ignored = {'P13': [[10, 20]]}
        add_pattern_if_ignored(pattern_ignored, pattern_item, results)
        self.assertEqual(results, [])

    def test_pattern_not_ignore(self):
        pattern_item = {'code_lines': [20, 30],
                        'pattern_code': 'P14',
                        'pattern_name': 'Null check',
                        'importance': 30.95612931128819}
        results = []
        pattern_ignored = {'P14': [[60, 100]]}
        add_pattern_if_ignored(pattern_ignored, pattern_item, results)
        self.assertEqual(results[0]['code_lines'], pattern_item['code_lines'])




import os
from unittest import TestCase
from aibolit.patterns.non_final_attribute.non_final_attribute import NonFinalAttribute
from pathlib import Path


class TestNonFinalAttribute(TestCase):
    dir_path = Path(os.path.realpath(__file__)).parent

    def test_find_non_final_atributes(self):
        lines = NonFinalAttribute().value(Path(self.dir_path, 'NonFinalAttribute.java'))
        self.assertEqual(lines, [2, 4, 6, 8])

    def test_nested_class(self):
        lines = NonFinalAttribute().value(Path(self.dir_path, 'File.java'))
        self.assertEqual(lines, [12, 16, 64])

    def test_attribute_in_interface(self):
        lines = NonFinalAttribute().value(Path(self.dir_path, 'AttributeInInterface.java'))
        self.assertEqual(lines, [])



from javalang.parse import parse
from javalang.tree import CompilationUnit

from aibolit.utils.encoding_detector import read_text_with_autodetected_encoding


def build_ast(filename: str) -> CompilationUnit:
    return parse(read_text_with_autodetected_encoding(filename))



from aibolit.utils.ast_builder import build_ast

from javalang.tree import MethodDeclaration, ClassDeclaration
from typing import Type


class MaxDiamOfTree:
    '''
    Returns max diameter for each method.
    This value is sum of Height of highest two subtree + 1
    '''
    def __init__(self):
        pass

    def depthOfTree(self, node: Type) -> int:
        '''
        Utility function that will return the depth of the tree
        '''
        maxdepth = 0

        node_arr = node if isinstance(node, list) else [node]
        for child in node_arr:
            if not hasattr(child, 'children'):
                return 0

            # Check for all children and find the maximum depth
            for each_child in child.children:
                maxdepth = max(maxdepth, self.depthOfTree(each_child))

        return maxdepth + 1

    def diameter(self, root: MethodDeclaration) -> int:
        '''
        Function to calculate the diameter of the tree
        '''
        # Find top two highest children
        max1 = 0
        max2 = 0
        root_arr = root if isinstance(root, list) else [root]
        for child in root_arr:
            if not hasattr(child, 'children'):
                return 0

            # Iterate over each child for diameter
            for each_child in child.children:
                h = self.depthOfTree(each_child)
                if h > max1:
                    max2 = max1
                    max1 = h
                elif h > max2:
                    max2 = h

        maxChildDia = 0
        for child in root_arr:
            if not hasattr(child, 'children'):
                return 0

            for each_child in child.children:
                maxChildDia = max(maxChildDia, self.diameter(each_child))

        return max(maxChildDia, max1 + max2 + 1)

    def value(self, filename: str):
        tree = build_ast(filename)

        traversed = []
        for _, class_body in tree.filter(ClassDeclaration):
            for each_object in class_body.body:
                if isinstance(each_object, MethodDeclaration):
                    traversed.append(self.diameter(each_object))

        return max(traversed) if len(traversed) != 0 else traversed



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import List
from networkx import DiGraph, dfs_labeled_edges  # type: ignore

from aibolit.types_decl import LineNumber
from aibolit.utils.ast_builder import build_ast
from aibolit.ast_framework import AST, ASTNodeType
from aibolit.utils.scope_status import ScopeStatus, ScopeStatusFlags


class VarMiddle:
    '''
    Returns lines in the file where variables declared in the middle
    of the method
    '''

    def value(self, filename):
        ast = AST.build_from_javalang(build_ast(filename))
        scope_status = ScopeStatus()
        lines_with_error: List[LineNumber] = []
        for _, destination, edge_type in dfs_labeled_edges(ast.tree, ast.root):
            if edge_type == 'forward':
                VarMiddle._on_entering_node(destination, ast.tree, scope_status, lines_with_error)
            elif edge_type == 'reverse':
                VarMiddle._on_leaving_node(destination, ast.tree, scope_status)

        return lines_with_error

    @staticmethod
    def _on_entering_node(node: int, ast: DiGraph, scope_status: ScopeStatus,
                          lines_with_error: List[LineNumber]):
        node_type = ast.nodes[node]['node_type']

        # if the variable is declared mark it and check the scope
        if node_type in VarMiddle._var_declaration_node_types:
            scope_status.add_flag(ScopeStatusFlags.INSIDE_VARIABLE_DECLARATION_SUBTREE)
            if ScopeStatusFlags.ONLY_VARIABLE_DECLARATIONS_PRESENT not in scope_status.get_status():
                lines_with_error.append(ast.nodes[node]['line'])

        # mark scope for super constructor calling
        elif node_type == ASTNodeType.STATEMENT_EXPRESSION:
            children_types = {ast.nodes[child]['node_type'] for child in ast.succ[node]}
            if ASTNodeType.SUPER_CONSTRUCTOR_INVOCATION in children_types:
                scope_status.add_flag(ScopeStatusFlags.INSIDE_CALLING_SUPER_CLASS_CONSTRUCTOR_SUBTREE)

        # mark scope for annotation usage
        elif node_type == ASTNodeType.ANNOTATION:
            scope_status.add_flag(ScopeStatusFlags.INSIDE_ANNOTATION_SUBTREE)

        else:
            # if we are not calling super constructor or declaring a variable
            # and node type not in black list spoil the scope
            if len(scope_status.get_status() & VarMiddle._ignore_scope_statuses) == 0 and \
               node_type not in VarMiddle._ignore_node_types:
                scope_status.remove_flag(ScopeStatusFlags.ONLY_VARIABLE_DECLARATIONS_PRESENT)

            if node_type in VarMiddle._new_scope_node_types:
                scope_status.enter_new_scope()

    @staticmethod
    def _on_leaving_node(node: int, ast: DiGraph, scope_status: ScopeStatus):
        node_type = ast.nodes[node]['node_type']

        # on the end of variable declaration remove according flag
        if node_type in VarMiddle._var_declaration_node_types:
            scope_status.remove_flag(ScopeStatusFlags.INSIDE_VARIABLE_DECLARATION_SUBTREE)

        # on the end of super constructor call remove according flag
        elif node_type == ASTNodeType.STATEMENT_EXPRESSION:
            children_types = {ast.nodes[child]['node_type'] for child in ast.succ[node]}
            if ASTNodeType.SUPER_CONSTRUCTOR_INVOCATION in children_types:
                scope_status.remove_flag(ScopeStatusFlags.INSIDE_CALLING_SUPER_CLASS_CONSTRUCTOR_SUBTREE)

        # on the end of annotation remove according flag
        elif node_type == ASTNodeType.ANNOTATION:
            scope_status.remove_flag(ScopeStatusFlags.INSIDE_ANNOTATION_SUBTREE)

        elif node_type in VarMiddle._new_scope_node_types:
            scope_status.leave_current_scope()

    _new_scope_node_types = {
        ASTNodeType.METHOD_DECLARATION,
        ASTNodeType.IF_STATEMENT,
        ASTNodeType.FOR_STATEMENT,
        ASTNodeType.SWITCH_STATEMENT,
        ASTNodeType.TRY_STATEMENT,
        ASTNodeType.DO_STATEMENT,
        ASTNodeType.WHILE_STATEMENT,
        ASTNodeType.BLOCK_STATEMENT,
        ASTNodeType.CATCH_CLAUSE,
        ASTNodeType.SYNCHRONIZED_STATEMENT,
    }

    _var_declaration_node_types = {
        ASTNodeType.LOCAL_VARIABLE_DECLARATION,
        ASTNodeType.TRY_RESOURCE,
    }

    _ignore_node_types = {
        ASTNodeType.FORMAL_PARAMETER,
        ASTNodeType.REFERENCE_TYPE,
        ASTNodeType.BASIC_TYPE,
        ASTNodeType.CATCH_CLAUSE_PARAMETER,
        ASTNodeType.ANNOTATION,
        ASTNodeType.TYPE_ARGUMENT,
        ASTNodeType.COLLECTION,
        ASTNodeType.STRING,
    }

    _ignore_scope_statuses = {
        ScopeStatusFlags.INSIDE_VARIABLE_DECLARATION_SUBTREE,
        ScopeStatusFlags.INSIDE_CALLING_SUPER_CLASS_CONSTRUCTOR_SUBTREE,
        ScopeStatusFlags.INSIDE_ANNOTATION_SUBTREE,
    }



import os
from unittest import TestCase
from pathlib import Path
from aibolit.patterns.protected_method.protected_method import ProtectedMethod


class TestProtectedMethod(TestCase):
    dir_path = Path(os.path.realpath(__file__)).parent

    def test_not_find_protected_method(self):
        self.assertEqual([], ProtectedMethod().value(Path(self.dir_path, 'NoProtectedMethod.java')),
                         'Should not match pattern protected method')

    def test_find_protected_method(self):
        self.assertEqual([2, 6], ProtectedMethod().value(Path(self.dir_path, 'ProtectedMethod.java')),
                         'Should match pattern protected method')

    def test_find_protected_method_inner(self):
        self.assertEqual([2, 11], ProtectedMethod().value(Path(self.dir_path, 'InnerClassProtectedMethod.java')),
                         'Should match pattern protected method in inner class')

    def test_find_protected_method_anonymous(self):
        self.assertEqual([5], ProtectedMethod().value(Path(self.dir_path, 'AnonymousClassProtectedMethod.java')),
                         'Should match pattern protected method in anonymous class')



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import List

from aibolit.ast_framework import AST, ASTNode, ASTNodeType
from aibolit.utils.ast_builder import build_ast


class CountIfReturn:
    '''
    Finds "if" statements with else branches and return statement in then branch.
    '''

    def value(self, filename: str) -> List[int]:
        ast = AST.build_from_javalang(build_ast(filename))
        lines: List[int] = []
        for if_statement in ast.get_proxy_nodes(ASTNodeType.IF_STATEMENT):
            if if_statement.else_statement is not None and \
               self._is_then_branch_return(if_statement):
                lines.append(if_statement.line)

        return lines

    def _is_then_branch_return(self, if_statement: ASTNode) -> bool:
        if if_statement.then_statement.node_type == ASTNodeType.RETURN_STATEMENT:
            return True

        if if_statement.then_statement.node_type == ASTNodeType.BLOCK_STATEMENT:
            return any(statement.node_type == ASTNodeType.RETURN_STATEMENT
                       for statement in if_statement.then_statement.statements)

        return False



from aibolit.utils.ast_builder import build_ast

from javalang.tree import VariableDeclarator, MemberReference


class NumVars:
    '''
    This class count number variables in class
    input: filename
    output: number of variables
    '''
    def __init__(self):
        pass

    def value(self, filename: str) -> int:
        tree = build_ast(filename)
        count_variables = set()
        for _, var_body in tree.filter(VariableDeclarator):
            count_variables.add(var_body.name)

        for _, var_body1 in tree.filter(MemberReference):
            count_variables.add(var_body1.member)

        return len(count_variables)



from typing import List
from aibolit.ast_framework import ASTNodeType, AST
from aibolit.utils.ast_builder import build_ast
from aibolit.ast_framework.ast_node import ASTNode


class ClassicSetter:
    '''
    The method's name starts with set, then goes
    the name of the attribute. There are attributes
    assigning in the method. Also, asserts are ignored.
    '''
    suitable_nodes: List[ASTNodeType] = [
        ASTNodeType.ASSERT_STATEMENT,
        ASTNodeType.STATEMENT_EXPRESSION,
    ]

    def _check_this_statements(self, ast: AST, node: ASTNode) -> bool:
        for child_this in ast.get_subtree(node).get_proxy_nodes(ASTNodeType.THIS):
            child_membref = child_this.selectors
            if len(child_membref):
                mem_referenced_name = child_membref[0].member
                if node.name.lower() == 'set' + mem_referenced_name.lower():
                    return True
        return False

    def _check_body_nodes(self, check_setter_body: List[ASTNode]) -> bool:
        '''
        Check whether nodes are agree with the following types
        (in self.suitable_nodes) or not.
        '''
        for node in check_setter_body:
            if node.node_type not in self.suitable_nodes:
                return False
        return True

    def value(self, filename: str) -> List[int]:
        lines: List[int] = []
        ast = AST.build_from_javalang(build_ast(filename))
        for node in ast.get_proxy_nodes(ASTNodeType.METHOD_DECLARATION):
            method_name = node.name
            if node.return_type is None and method_name.startswith('set') and \
               self._check_body_nodes(node.body) and self._check_this_statements(ast, node):
                lines.append(node.line)
        return sorted(lines)



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
from unittest import TestCase
from aibolit.patterns.string_concat.string_concat import StringConcatFinder
from pathlib import Path


class TestConcatString(TestCase):
    dir_path = Path(os.path.realpath(__file__)).parent
    concat_finder = StringConcatFinder()

    def test_concat_strings_in_print(self):
        lines = self.concat_finder.value(Path(self.dir_path, 'ConcatInPrint.java'))
        self.assertEqual(lines, [14])

    def test_member_plus_string(self):
        lines = self.concat_finder.value(Path(self.dir_path, 'MemberPlusString.java'))
        self.assertEqual(lines, [99, 111])

    def test_empty_case(self):
        lines = self.concat_finder.value(Path(self.dir_path, 'Nothing.java'))
        self.assertEqual(lines, [])

    def test_string_plus_member(self):
        lines = self.concat_finder.value(Path(self.dir_path, 'StringPlusMember.java'))
        self.assertEqual(len(lines), 1)

    def test_many_concats(self):
        lines = self.concat_finder.value(Path(self.dir_path, 'ManyConcats.java'))
        self.assertEqual(lines, [12, 13, 14, 15])

    def test_concat_in_different_methods(self):
        lines = self.concat_finder.value(Path(self.dir_path, 'DifferentMethods.java'))
        self.assertEqual(lines, [13, 27])

    def test_fake_operator_plus(self):
        lines = self.concat_finder.value(Path(self.dir_path, 'FakePlusOperator.java'))
        self.assertEqual(lines, [])

    def test_string_with_quotes(self):
        lines = self.concat_finder.value(Path(self.dir_path, 'RustServerCodegen.java'))
        self.assertEqual(lines, [
            352, 372, 373, 379, 380, 408, 409, 418, 422, 429, 430, 438, 456, 466,
            479, 494, 499, 509, 562, 563, 604, 605, 610, 701, 995, 1009, 1012, 1029,
            1032, 1080, 1258, 1267, 1337, 1386, 1414
        ])

    def test_comment_inside_line(self):
        lines = self.concat_finder.value(Path(self.dir_path, 'XMLDataObject.java'))
        self.assertEqual(lines, [160, 165, 177, 192, 218, 279])

    def test_fake1(self):
        lines = self.concat_finder.value(Path(self.dir_path, 'Chain.java'))
        self.assertEqual(lines, [32])



import javalang
import re
import itertools

from aibolit.utils.ast_builder import build_ast


class VarSiblings:
    def __init__(self):
        pass

    def value(self, filename: str):
        numbers = []
        for first, second in itertools.product(
            [node
                for _, node in build_ast(filename).filter(javalang.tree.LocalVariableDeclaration)
                if len(re.findall(r'[A-Z]?[a-z]+|[A-Z]+(?=[A-Z]|$)', node.declarators[0].name)) > 1
                and len(re.findall(r'[A-Z]?[a-z]+|[A-Z]+(?=[A-Z]|$)', node.declarators[0].name)[0]) > 3],  # noqa: W503
                repeat=2
        ):
            if first != second:
                if second.declarators[0].name.startswith(
                    re.findall(r'[A-Z]?[a-z]+|[A-Z]+(?=[A-Z]|$)', first.declarators[0].name)[0]
                ):
                    numbers.append(first.position.line)
        return numbers



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import subprocess
import os
import shutil
from bs4 import BeautifulSoup
import tempfile
import uuid


class CCMetric():
    """Main Cyclical Complexity class."""

    input = ''

    def __init__(self, input):
        """Initialize class."""
        if len(input) == 0:
            raise ValueError('Empty file for analysis')
        self.input = input

    def value(self, showoutput=False):
        """Run Cyclical Complexity analaysis"""
        try:
            root = os.path.join(tempfile.gettempdir(), uuid.uuid4().hex)
            dirName = os.path.join(root, 'src/main/java')
            os.makedirs(dirName)
            if os.path.isdir(self.input):
                shutil.copytree(self.input, os.path.join(dirName, self.input))
            elif os.path.isfile(self.input):
                pos1 = self.input.rfind('/')
                os.makedirs(dirName + '/' + self.input[0:pos1])
                shutil.copyfile(self.input, os.path.join(dirName, self.input))
            else:
                raise Exception('File ' + self.input + ' does not exist')
            tmppath = os.path.dirname(os.path.realpath(__file__))
            shutil.copyfile(os.path.join(tmppath, 'pom.xml'), root + '/pom.xml')
            shutil.copyfile(os.path.join(tmppath, 'cyclical.xml'), root + '/cyclical.xml')
            if showoutput:
                subprocess.run(['mvn', 'pmd:pmd'], cwd=root)
            else:
                subprocess.run(['mvn', 'pmd:pmd'], cwd=root,
                               stdout=subprocess.DEVNULL,
                               stderr=subprocess.DEVNULL)
            if os.path.isfile(root + '/target/pmd.xml'):
                res = self.__parseFile(root)
                return res
            raise Exception('File ' + self.input + ' analyze failed')
        finally:
            shutil.rmtree(root)

    def __parseFile(self, root):
        result = {'data': [], 'errors': []}
        content = []
        with open(root + '/target/pmd.xml', 'r') as file:
            content = file.read()
            soup = BeautifulSoup(content, 'lxml')
            files = soup.find_all("file")
            for file in files:
                out = file.violation.string
                name = file['name']
                pos1 = name.find(root + '/src/main/java/')
                pos1 = pos1 + len(root + '/src/main/java/')
                name = name[pos1:]
                pos1 = out.find('has a total cyclomatic complexity')
                pos1 = out.find('of ', pos1)
                pos1 = pos1 + 3
                pos2 = out.find('(', pos1)
                complexity = int(out[pos1:pos2 - 1])
                result['data'].append({'file': name, 'complexity': complexity})
            errors = soup.find_all("error")
            for error in errors:
                name = error['filename']
                pos1 = name.find(root + '/src/main/java/')
                pos1 = pos1 + len(root + '/src/main/java/')
                name = name[pos1:]
                raise Exception(error['msg'])
        return result



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
from unittest import TestCase
from aibolit.patterns.return_null.return_null import ReturnNull
from pathlib import Path


class TestMethodChain(TestCase):
    dir_path = Path(os.path.realpath(__file__)).parent
    method_chain_finder = ReturnNull()

    def test_anonymous(self):
        lines = self.method_chain_finder.value(Path(self.dir_path, 'Anonymous.java'))
        self.assertEqual(lines, [24, 28])

    def test_empty(self):
        lines = self.method_chain_finder.value(Path(self.dir_path, 'Empty.java'))
        self.assertEqual(lines, [])

    def test_with_ternary1(self):
        lines = self.method_chain_finder.value(Path(self.dir_path, 'With_Ternary1.java'))
        self.assertEqual(lines, [12])

    def test_with_ternary2(self):
        lines = self.method_chain_finder.value(Path(self.dir_path, 'With_Ternary2.java'))
        self.assertEqual(lines, [12])

    def test_with_ternary_not_return_null(self):
        lines = self.method_chain_finder.value(Path(self.dir_path, 'With_Ternary_not_return_null.java'))
        self.assertEqual(lines, [])

    def test_simple(self):
        lines = self.method_chain_finder.value(Path(self.dir_path, 'Simple.java'))
        self.assertEqual(lines, [12])




import os
from unittest import TestCase
from aibolit.patterns.er_class.er_class import ErClass
from pathlib import Path


class TestErClass(TestCase):
    dir_path = Path(os.path.realpath(__file__)).parent
    er_class_finder = ErClass()

    def test_manager_in_middle(self):
        lines = self.er_class_finder.value(Path(self.dir_path, 'AnimatableSplitDimensionPathValue.java'))
        self.assertEqual(lines, [12])

    def test_controller_in_end(self):
        lines = self.er_class_finder.value(Path(self.dir_path, 'AnimatableTransform.java'))
        self.assertEqual(lines, [12])

    def test_one_normal_class(self):
        lines = self.er_class_finder.value(Path(self.dir_path, 'AuditEventModelProcessor.java'))
        self.assertEqual(lines, [])

    def test_two_classes_with_pattern(self):
        lines = self.er_class_finder.value(Path(self.dir_path, 'BaseKeyframeAnimation.java'))
        self.assertEqual(lines, [18, 186])

    def test_class_parser(self):
        lines = self.er_class_finder.value(Path(self.dir_path, 'Configuration.java'))
        self.assertEqual(lines, [3106])

    def test_another_normal_class(self):
        lines = self.er_class_finder.value(Path(self.dir_path, 'FillContent.java'))
        self.assertEqual(lines, [])

    def test_four_normal_classes(self):
        lines = self.er_class_finder.value(Path(self.dir_path, 'FJIterateTest.java'))
        self.assertEqual(lines, [])

    def test_two_distant_normal_classes(self):
        lines = self.er_class_finder.value(Path(self.dir_path, 'FJListProcedureRunner.java'))
        self.assertEqual(lines, [])

    def test_classes_in_comments(self):
        lines = self.er_class_finder.value(Path(self.dir_path, 'KeyProviderCryptoExtension.java'))
        self.assertEqual(lines, [])

    def test_classes_in_methods(self):
        lines = self.er_class_finder.value(Path(self.dir_path, 'OsSecureRandom.java'))
        self.assertEqual(lines, [])

    def test_normal_class(self):
        lines = self.er_class_finder.value(Path(self.dir_path, 'RectangleContent.java'))
        self.assertEqual(lines, [])

    def test_three_writers_one_reader(self):
        lines = self.er_class_finder.value(Path(self.dir_path, 'SequenceFile.java'))
        self.assertEqual(lines, [837, 1478, 1538, 1684])



#!/usr/bin/env python
# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""The main entry point. Invoke as `aibolit' or `python -m aibolit'.
"""

import argparse
import concurrent.futures
import json
import multiprocessing
import operator
import os
import pickle
import sys
import time
import traceback
from collections import defaultdict
from os import scandir
from pathlib import Path
from sys import stdout
from typing import List, Any, Dict, Tuple

import javalang
import numpy as np  # type: ignore
import requests
from lxml import etree  # type: ignore
from pkg_resources import parse_version

from aibolit import __version__
from aibolit.config import Config
from aibolit.ml_pipeline.ml_pipeline import train_process, collect_dataset
from aibolit.utils.ast_builder import build_ast

dir_path = os.path.dirname(os.path.realpath(__file__))


def list_dir(path, files):
    dir_list = []
    for entry in scandir(path):
        if entry.is_dir():
            dir_list.append(entry.path)
            dir_list.extend(list_dir(entry.path, files))
        elif entry.is_file() \
                and entry.path.endswith('.java') \
                and entry.name != 'package-info.java':
            files.append(entry.path)
    return dir_list


def predict(input_params, model, args):
    features_order = model.features_conf['features_order']
    # add ncss to last column. We will normalize all patterns by that value
    # deepcode ignore ExpectsIntDislikesStr: false-positive recommendation of deepcode
    input = [input_params[i] for i in features_order] + [input_params['M2']]
    th = float(args.threshold) or 1.0
    preds, importances = model.informative(input, th=th)

    return {features_order[int(x)]: int(x) for x in preds}, list(importances)


def run_parse_args(commands_dict):
    parser = argparse.ArgumentParser(
        description='Find the pattern which has the largest impact on readability',
        usage='''aibolit <command> [<args>]

        You can run 1 command:
        train          Train model
        check          Recommend pattern
        recommend      Recommend pattern. The same as recommend, just another acronym''')

    parser.add_argument('command', help='Subcommand to run')
    parser.add_argument(
        '--version', action='version',
        version='%(prog)s {version}'.format(version=__version__)
    )
    # parse_args defaults to [1:] for args, but you need to
    # exclude the rest of the args too, or validation will fail
    args = parser.parse_args(sys.argv[1:2])
    if args.command not in commands_dict:
        parser.print_help()
        exit(1)

    return commands_dict[args.command]()


def train():
    parser = argparse.ArgumentParser(
        description='Collect dataset and train model'
    )
    parser.add_argument(
        '--java_folder',
        help='full path to directory',
        default=False,
        required=False
    )
    parser.add_argument(
        '--max_classes',
        type=lambda v: sys.maxsize if v == '' else int(v),
        required=False,
        default=None
    )
    parser.add_argument(
        '--skip_collect_dataset',
        required=False,
        default=False,
        action='store_true'
    )
    parser.add_argument(
        '--dataset_file',
        required=False,
        default=False
    )
    args = parser.parse_args(sys.argv[2:])
    if not args.skip_collect_dataset:
        collect_dataset(args)
    train_process()


def __count_value(value_dict, input_params, code_lines_dict, java_file: str, is_metric=False):
    """
    Count value for input dict

    :param value_dict: Pattern item or Metric item from CONFIG
    :param input_params: list with calculated patterns/metrics
    :param code_lines_dict: list with found code lines of patterns/metrics
    :param java_file: full path for java file
    :is_metric: is item metric?
    :return: None, it has side-effect
    """
    acronym = value_dict['code']
    try:
        val = value_dict['make']().value(java_file)
        if not is_metric:
            input_params[acronym] = len(val)
            code_lines_dict['lines_' + acronym] = val
        else:
            input_params[acronym] = val
    except Exception:
        exc_type, exc_value, exc_tb = sys.exc_info()
        raise Exception("Can't count {} metric: {}".format(
            acronym,
            str(type(exc_value)))
        )


def flatten(l):
    return [item for sublist in l for item in sublist]


def add_pattern_if_ignored(
        dct: Dict[str, Any],
        pattern_item: Dict[Any, Any],
        results_list: List[Any]) -> None:
    """ If pattern code is not ignored, add it to the result list

    :param dct: dict, where key is pattern, value is list of lines range to ignore
    :param pattern_item: pattern dict which was get after `inference` function
    :param results_list: result list to add

    """
    ignored_lines = dct.get(pattern_item['pattern_code'])
    if ignored_lines:
        for place in ignored_lines:
            # get lines range of ignored code
            start_line_to_ignore = place[0]
            end_line_to_ignore = place[1]
            new_code_lines = []
            for line in pattern_item['code_lines']:
                if (line >= start_line_to_ignore) and (line <= end_line_to_ignore):
                    continue
                else:
                    new_code_lines.append(line)
            pattern_item['code_lines'] = new_code_lines
            if len(pattern_item['code_lines']) > 0:
                results_list.append(pattern_item)
    else:
        results_list.append(pattern_item)


def find_annotation_by_node_type(
        tree: javalang.tree.CompilationUnit,
        node_type) -> Dict[Any, Any]:
    """Search nodes with annotations.

    :param tree: javalang.tree
    :param node_type: Node type of javalang.tree
    :return
    dict with annotations, where key is node, value is list of string annotations;
    """
    annonations = defaultdict(list)
    for _, node in tree.filter(node_type):
        if node.annotations:
            for a in node.annotations:
                if hasattr(a.element, 'value'):
                    if 'aibolit' in a.element.value:
                        annonations[node].append(
                            a.element.value.split('.')[1].rstrip('\"')
                        )
                elif hasattr(a.element, 'values'):
                    for j in a.element.values:
                        if 'aibolit' in j.value:
                            annonations[node].append(
                                j.value.split('.')[1].rstrip('\"')
                            )
    return annonations


def find_start_and_end_lines(node) -> Tuple[int, int]:  # noqa: C901
    max_line = node.position.line

    def check_max_position(node):
        nonlocal max_line
        if hasattr(node, '_position'):
            if node.position.line > max_line:
                max_line = node.position.line

    def traverse(node):
        nonlocal max_line
        check_max_position(node)

        if hasattr(node, 'children'):
            for child in node.children:
                if isinstance(child, list) and (len(child) > 0):
                    for item in child:
                        traverse(item)
                else:
                    if hasattr(child, 'children'):
                        infants = child.children
                        for infant in infants:
                            traverse(infant)
                    else:
                        check_max_position(child)

        else:
            return

    traverse(node)
    return node.position.line, max_line


def calculate_patterns_and_metrics(file, args):
    code_lines_dict = input_params = {}  # type: ignore
    error_string = None
    patterns_to_suppress = args.suppress
    try:
        config = Config.get_patterns_config()
        for pattern in config['patterns']:
            if pattern['code'] in config['patterns_exclude']:
                continue
            if pattern['code'] in patterns_to_suppress:
                input_params[pattern['code']] = 0
            else:
                __count_value(pattern, input_params, code_lines_dict, file)

        for metric in config['metrics']:
            if metric['code'] in config['metrics_exclude']:
                continue
            __count_value(metric, input_params, code_lines_dict, file, is_metric=True)
    except Exception as ex:
        error_string = str(ex)
        input_params = []  # type: ignore

    return input_params, code_lines_dict, error_string


def inference(
        input_params: List[int],
        code_lines_dict,
        args):
    """
    Find a pattern which has the largest impact on target

    :param input_params: list if calculated patterns/metrics
    :param code_lines_dict: list with found code lines of patterns/metrics
    :param file: filename

    :return: list of results with code_lies for each pattern and its name
    """
    model_path = args.model
    do_full_report = args.full
    results = []
    if input_params:
        if not model_path:
            model_path = Config.folder_model_data()
        with open(model_path, 'rb') as fid:
            model = pickle.load(fid)
        sorted_result, importances = predict(input_params, model, args)
        patterns_list = model.features_conf['patterns_only']
        for iter, (key, val) in enumerate(sorted_result.items()):
            if key in patterns_list:
                pattern_code = key
                code_lines = code_lines_dict.get('lines_' + key)
                importance = importances[iter] * input_params[pattern_code]

                # We show only patterns with positive importance
                if code_lines and importance > 0:
                    if code_lines:
                        pattern_name = \
                            [x['name'] for x in Config.get_patterns_config()['patterns']
                             if x['code'] == pattern_code][0]
                        results.append(
                            {'code_lines': code_lines,
                             'pattern_code': pattern_code,
                             'pattern_name': pattern_name,
                             'importance': importance
                             })
                    if not do_full_report:
                        break
    else:
        return results

    return results


def run_recommend_for_file(file: str, args):
    """
    Calculate patterns and metrics, pass values to model and suggest pattern to change
    :param file: file to analyze
    :param args: different command line arguments
    :return: dict with code lines, filename and pattern name
    """
    java_file = str(Path(os.getcwd(), file))
    tree = build_ast(file)
    classes_with_annonations = find_annotation_by_node_type(tree, javalang.tree.ClassDeclaration)
    functions_with_annotations = find_annotation_by_node_type(tree, javalang.tree.MethodDeclaration)
    fields_with_annotations = find_annotation_by_node_type(tree, javalang.tree.FieldDeclaration)
    classes_with_patterns_ignored = flatten(
        [pattern_code for node, pattern_code in classes_with_annonations.items()])
    patterns_ignored = defaultdict(list)

    for node, patterns_list in functions_with_annotations.items():
        start_pos, end_pos = find_start_and_end_lines(node)
        for p in patterns_list:
            patterns_ignored[p].append([start_pos, end_pos])

    for node, patterns_list in fields_with_annotations.items():
        for p in patterns_list:
            patterns_ignored[p].append([node.position.line, node.position.line])

    input_params, code_lines_dict, error_string = calculate_patterns_and_metrics(java_file, args)

    if not input_params:
        results_list = []  # type: ignore
        error_string = 'Empty java file; ncss = 0'
    #  deepcode ignore ExpectsIntDislikesStr: False positive
    elif input_params['M2'] == 0:
        results_list = []  # type: ignore
        error_string = 'Empty java file; ncss = 0'
    else:
        results_list = inference(input_params, code_lines_dict, args)
        new_results: List[Any] = []
        for pattern_item in results_list:
            # check if the whole class is suppressed
            if pattern_item['pattern_code'] not in classes_with_patterns_ignored:
                # then check if patterns are ignored in fields or functions
                add_pattern_if_ignored(patterns_ignored, pattern_item, new_results)
                # add_pattern_if_ignored(patterns_for_fields_ignored, pattern_item, new_results)
            else:
                continue

        results_list = new_results

    if error_string:
        ncss = 0
    else:
        ncss = input_params.get('M4', 0)

    return {
        'filename': file,
        'results': results_list,
        'error_string': error_string,
        'ncss': ncss,
    }


def create_xml_tree(results, full_report, cmd, exit_code):
    """
    Creates xml from output of `check` function
    :param results: output of `check` function
    :return: xml string
    """
    importances_for_all_classes = []
    top = etree.Element('report')
    header_tag = etree.SubElement(top, 'header')
    importances_for_all_classes_tag = etree.SubElement(header_tag, 'score')
    datetime_tag = etree.SubElement(header_tag, 'datetime')
    datetime_tag.addprevious(etree.Comment('datetime in ms'))
    datetime_tag.text = str(int(round(time.time() * 1000)))
    version_tag = etree.SubElement(header_tag, 'version')
    version_tag.text = str(__version__)
    files_number_tag = etree.SubElement(header_tag, 'files')
    files_number_tag.addprevious(etree.Comment('Files with found patterns'))
    files_number_tag.text = str(len(
        [x for x in results
         if (not x.get('error_string') and x.get('results'))]
    ))
    patterns_number_tag = etree.SubElement(header_tag, 'patterns')
    ncss_tag = etree.SubElement(header_tag, 'ncss')
    ncss_tag.text = str(sum([x['ncss'] for x in results]))
    cmd_tag = etree.SubElement(header_tag, 'cmd')
    cmd_tag.text = ' '.join(cmd)
    cmd_tag = etree.SubElement(header_tag, 'exit_code')
    cmd_tag.text = str(exit_code)
    total_patterns = 0
    files = etree.SubElement(top, 'files')
    if not full_report:
        files.addprevious(etree.Comment('Show pattern with the largest contribution to Cognitive Complexity'))
    else:
        files.addprevious(etree.Comment('Show all patterns'))
    for result_for_file in results:
        file = etree.SubElement(files, 'file')
        filename = result_for_file.get('filename')
        name = etree.SubElement(file, 'path')
        output_string_tag = etree.SubElement(file, 'summary')
        name.text = filename
        results_item = result_for_file.get('results')
        errors_string = result_for_file.get('error_string')
        if not results_item and not errors_string:
            output_string = 'Your code is perfect in aibolit\'s opinion'
            output_string_tag.text = output_string
        elif not results_item and errors_string:
            output_string = 'Error when calculating patterns: {}'.format(str(errors_string))
            output_string_tag.text = output_string
        else:
            output_string = 'Some issues found'
            output_string_tag.text = output_string
            importances_sum_tag = etree.SubElement(file, 'score')
            patterns_tag = etree.SubElement(file, 'patterns')
            patterns_number = len(result_for_file['results'])
            total_patterns += patterns_number
            importance_for_class = []
            for i, pattern in enumerate(result_for_file['results'], start=1):
                if pattern.get('pattern_code'):
                    pattern_item = etree.SubElement(patterns_tag, 'pattern')
                    pattern_name_str = pattern.get('pattern_name')
                    details = etree.SubElement(pattern_item, 'details')
                    details.text = pattern_name_str or ''
                    pattern_item.attrib['code'] = pattern.get('pattern_code')
                    code_lines_items = pattern.get('code_lines')
                    pattern_score = pattern.get('importance')
                    pattern_score_tag = etree.SubElement(pattern_item, 'score')
                    pattern_score_tag.text = '{:.2f}'.format(pattern_score) or ''
                    pattern_score_tag = etree.SubElement(pattern_item, 'order')
                    pattern_score_tag.text = '{}/{}'.format(i, patterns_number)
                    importance_for_class.append(pattern_score)
                    if code_lines_items:
                        code_lines_lst_tree_node = etree.SubElement(pattern_item, 'lines')
                        for code_line in code_lines_items:
                            code_line_elem = etree.SubElement(code_lines_lst_tree_node, 'number')
                            code_line_elem.text = str(code_line)
            score_for_class = sum(importance_for_class)
            importances_sum_tag.text = str(score_for_class)
            importances_for_all_classes.append(score_for_class)

    patterns_number_tag.text = str(total_patterns)
    if importances_for_all_classes:
        importances_for_all_classes_tag.text = str(np.mean(importances_for_all_classes))

    return top


def get_exit_code(results):
    """
    Analyzed recommendation results and generate exit_code for pipeline
    """

    files_analyzed = len(results)
    errors_number = 0
    perfect_code_number = 0
    for result_for_file in results:
        results = result_for_file.get('results')
        errors_string = result_for_file.get('error_string')
        if not results and not errors_string:
            perfect_code_number += 1
        elif not results and errors_string:
            errors_number += 1

    if errors_number == files_analyzed:
        # we have errors everywhere
        exit_code = 2
    elif perfect_code_number == files_analyzed:
        # everything is good
        exit_code = 0
    else:
        # we have some recommendation
        exit_code = 1

    return exit_code


def create_text(results, full_report, is_long=False):
    importances_for_all_classes = []
    buffer = []
    total_patterns = 0
    if not full_report:
        buffer.append('Show pattern with the largest contribution to Cognitive Complexity')
    else:
        buffer.append('Show all patterns')
    for result_for_file in results:
        filename = result_for_file.get('filename')
        results_item = result_for_file.get('results')
        errors_string = result_for_file.get('error_string')
        if not results_item and not errors_string:
            # Do nothing, patterns were not found
            pass
        if not results_item and errors_string:
            output_string = '{}: error when calculating patterns: {}'.format(
                filename,
                str(errors_string)
            )
            buffer.append(output_string)
        elif results_item and not errors_string:
            # get unique patterns score
            patterns_scores = print_total_score_for_file(buffer, filename, importances_for_all_classes, result_for_file)
            patterns_number = len(patterns_scores)
            pattern_number = 0
            cur_pattern_name = ''
            for pattern_item in result_for_file['results']:
                pattern_score = pattern_item.get('importance')
                code = pattern_item.get('pattern_code')
                if code:
                    pattern_name_str = pattern_item.get('pattern_name')
                    if cur_pattern_name != pattern_name_str:
                        pattern_number += 1
                        cur_pattern_name = pattern_name_str
                    buffer.append('{}[{}]: {} ({}: {:.2f} {}/{})'.format(
                        filename,
                        pattern_item.get('code_line'),
                        pattern_name_str,
                        code,
                        pattern_score,
                        pattern_number,
                        patterns_number))

            total_patterns += pattern_number
    if importances_for_all_classes:
        show_summary(buffer, importances_for_all_classes, is_long, results, total_patterns)

    return buffer


def show_summary(buffer, importances_for_all_classes, is_long, results, total_patterns):
    files_number = len(
        [x for x in results
         if (not x.get('errors_string') and x.get('results'))
         ])
    ncss = sum(
        [x['ncss'] for x in results
         if (not x.get('errors_string') and x.get('results'))])
    if not is_long:
        buffer.append('Total score: {:.2f}, files seen: {}, patterns found: {}, ncss: {}'.format(
            np.mean(importances_for_all_classes),
            files_number,
            total_patterns,
            ncss
        ))
    else:
        buffer.append('Total score: {:.2f}, files seen: {}, ncss {}'.format(
            np.mean(importances_for_all_classes),
            files_number,
            ncss
        ))


def print_total_score_for_file(
        buffer: List[str],
        filename: str,
        importances_for_all_classes: List[int],
        result_for_file):
    patterns_scores = {}
    for x in result_for_file['results']:
        patterns_scores[x['pattern_name']] = x['importance']
    importances_for_class = sum(patterns_scores.values())
    importances_for_all_classes.append(importances_for_class)
    buffer.append('{} score: {}'.format(filename, importances_for_class))
    return patterns_scores


def check():
    """Run check pipeline."""

    parser = argparse.ArgumentParser(
        description='Get recommendations for Java code',
        usage='aibolit check < --folder | --filenames > [--model] '
              '[--threshold] [--full] [--format]')

    group_exclusive = parser.add_mutually_exclusive_group(required=True)

    group_exclusive.add_argument(
        '--folder',
        help='path to Java files',
        default=False
    )
    group_exclusive.add_argument(
        '--filenames',
        help='list of Java files',
        nargs="*",
        default=False
    )
    parser.add_argument(
        '--model',
        help='''file where pretrained model is located, the default path is located
        in site-packages and is installed with aibolit automatically''',
        default=False
    )

    parser.add_argument(
        '--threshold',
        help='threshold for predict',
        default=False
    )
    parser.add_argument(
        '--full',
        help='show all recommendations instead of the best one',
        default=False,
        action='store_true'
    )
    parser.add_argument(
        '--format',
        default='compact',
        help='compact (by default), long or xml. Usage: --format=xml'
    )

    parser.add_argument(
        '--suppress',
        default=[]
    )

    parser.add_argument(
        '--exclude',
        action='append',
        nargs='+'
    )

    args = parser.parse_args(sys.argv[2:])

    if args.suppress:
        args.suppress = args.suppress.strip().split(',')
    if args.threshold:
        print('Threshold for model has been set to {}'.format(args.threshold))

    files_to_exclude = handle_exclude_command_line(args)

    if args.filenames:
        files = [str(Path(x).absolute()) for x in args.filenames if x not in files_to_exclude]
    elif args.folder:
        all_files = []
        list_dir(args.folder, all_files)
        files = [str(Path(x).absolute()) for x in all_files if str(x) not in files_to_exclude]

    results = list(run_thread(files, args))
    exit_code = get_exit_code(results)

    if args.format:
        if args.format == 'xml':
            root = create_xml_tree(results, args.full, sys.argv, exit_code)
            tree = root.getroottree()
            tree.write(stdout.buffer, pretty_print=True)
        else:
            if args.format in ['compact', 'short']:
                new_results = format_converter_for_pattern(results)
            elif args.format == 'long':
                new_results = format_converter_for_pattern(results, 'code_line')
            else:
                raise Exception('Unknown format')
            text = create_text(new_results, args.full, True)

            if args.format == 'short':
                print(text[-1])
            else:
                print('\n'.join(text))

    return exit_code


def handle_exclude_command_line(args):
    files_to_exclude = []
    exc_string = 'Usage: --exclude=<glob pattern> ' \
                 '--exclude=<glob pattern> ... ' \
                 '--exclude=folder_to_find_exceptions '
    if args.exclude:
        if len(args.exclude) < 2:
            raise Exception(exc_string)
        try:
            folder_to_exclude = args.exclude[-1][0]
            glob_patterns = [x[0] for x in args.exclude[:-1]]
            for glob_p in glob_patterns:
                files_to_exclude.extend([str(x.absolute()) for x in list(Path(folder_to_exclude).glob(glob_p))])

        except Exception:
            raise Exception(exc_string)
    return files_to_exclude


def format_converter_for_pattern(results, sorted_by=None):
    """
    Reformat data where data are sorted by patterns importance
    (it is already sorted in the input).
    Then lines are sorted in ascending order.
    """

    for file in results:
        items = file.get('results')
        if items:
            new_items = flatten([
                [{'pattern_code': x['pattern_code'],
                  'pattern_name': x['pattern_name'],
                  'code_line': line,
                  'importance': x['importance']
                  } for line in sorted(x['code_lines'])] for x in items
            ])
            if not sorted_by:
                file['results'] = sorted(new_items, key=lambda e: (-e['importance'], e['pattern_code'], e['code_line']))
            else:
                file['results'] = sorted(new_items, key=operator.itemgetter(sorted_by, 'pattern_code'))

    return results


def version():
    """
    Parses arguments and shows current version of program.
    """

    parser = argparse.ArgumentParser(
        description='Show version')
    parser.add_argument(
        '--version',
    )
    print('%(prog)s {version}'.format(version=__version__))


def run_thread(files, args):
    """
    Parallel patterns/metrics calculation
    :param files: list of java files to analyze

    """
    with concurrent.futures.ThreadPoolExecutor(max_workers=multiprocessing.cpu_count()) as executor:
        future_results = [executor.submit(run_recommend_for_file, file, args) for file in files]
        concurrent.futures.wait(future_results)
        for future in future_results:
            yield future.result()


def get_versions(pkg_name):
    url = f'https://pypi.python.org/pypi/{pkg_name}/json'
    releases = json.loads(requests.get(url, timeout=1).content)['releases']
    return sorted(releases, key=parse_version, reverse=True)


def main():
    try:
        max_available_version = get_versions('aibolit')[0]
        if max_available_version != __version__:
            print('Version {} is available, but you are using {}'.format(
                max_available_version,
                __version__
            ))
    except requests.exceptions.ConnectionError:
        print('Can\'t check aibolit version. Network is not available')
    try:
        commands = {
            'train': lambda: train(),
            'check': lambda: check(),
            'recommend': lambda: check(),
            'version': lambda: version(),
        }
        exit_code = run_parse_args(commands)
    except Exception:
        traceback.print_exc()
        sys.exit(2)
    else:
        sys.exit(exit_code)


if __name__ == '__main__':
    main()
